import unittest
import os
from rickled import BaseRickle, Rickle

class TestAdvanced(unittest.TestCase):

    def test_load_lambda_overload(self):
        yaml = """
main:
    file:
        type: from_file
        file_path: './tests/placebos/test_security.yaml'
        load_as_rick: true
        deep: true
        load_lambda: false
        """

        # Unsafe loading
        test_conf_yaml_unsafe = Rickle(yaml, load_lambda=False)

        self.assertTrue(callable(test_conf_yaml_unsafe.main.file.root.link_to_file.bad_function))

        # Set env var and safe load
        os.environ["RICKLE_SAFE_LOAD"] = "1"

        test_conf_yaml_safe = Rickle(yaml, load_lambda=False)

        self.assertIsInstance(test_conf_yaml_safe.main.file.root.link_to_file.bad_function, dict)

        # Still safe due to env
        test_conf_yaml = Rickle('./tests/placebos/test_malicious.yaml', load_lambda=True)

        self.assertIsInstance(test_conf_yaml.bad_function, dict)

        # Set unsafe again
        del os.environ['RICKLE_SAFE_LOAD']

        test_conf_yaml = Rickle('./tests/placebos/test_malicious.yaml', load_lambda=True)

        self.assertTrue(callable(test_conf_yaml.bad_function))

    def test_search_path(self):
        yaml = """
path:
    to:
        second: 1
    and:
        second: 1
    another: 0
route:
    second: 1
        """

        test_rickle = BaseRickle(yaml)

        list_of_paths = test_rickle.search_path(key='second')
        expected = ['/path/to/second', '/path/and/second', '/route/second']

        self.assertIsInstance(list_of_paths, list)
        self.assertListEqual(list_of_paths, expected)

        list_of_paths = test_rickle.search_path(key='third')

        self.assertIsInstance(list_of_paths, list)
        self.assertEqual(len(list_of_paths), 0)

        for path in expected:
            self.assertEqual(test_rickle(path=path), 1)

    def test_path_string(self):

        yaml = """
path:
    datenow:
        type: lambda
        import:
            - "from datetime import datetime as dd"
        load: "dd.utcnow().strftime('%Y-%m-%d')"
    level_one:
        level_two:                    
            member: 42                    
            list_member:
                - 1
                - 0
                - 1 
                - 1
                - 1
        funcs:
            type: function
            name: funcs
            args:
                x: 42
                y: worl
            load: >
                def funcs(x, y):
                    _x = int(x) 
                    return f'Hello {y}, {_x / len(y)}!'
        """

        test_rickle = Rickle(yaml, load_lambda=True)

        v = test_rickle('/path/level_one/level_two/member')

        self.assertEqual(v, 42)

        v = test_rickle("/path/level_one/funcs?x=100&y='world'")

        self.assertEqual(v, 'Hello world, 20.0!')

        with self.assertRaises(NameError):
            v = test_rickle('/path/level_two/funcs?x=100&y=world')

        with self.assertRaises(TypeError):
            v = test_rickle('/path/level_one/level_two/member?x=100&y=world')

        with self.assertRaises(TypeError):
            v = test_rickle('/path/level_one/funcs?x=100&z=world')

        v = test_rickle('/path/level_one/funcs?x=100')

        self.assertEqual(v, 'Hello worl, 25.0!')

        v = test_rickle('/path/datenow')

        self.assertIsInstance(v, str)

        v = test_rickle('/path/level_one')

        self.assertIsInstance(v, Rickle)

        self.assertTrue(callable(v.funcs))

        self.assertIsInstance(v.level_two.list_member, list)

    def test_self_reference(self):
        y = """
const:
  f: 0.5
get_area:
  type: function
  name: get_area
  is_method: true
  args:
     x: 10
     y: 10
     z: null
  import:
     - math
  load: >
     def get_area(self, x, y, z):
        if not z is None:
           area = (x * y) + (x * z) + (y * z)
           area = 2 * area
        else:
           area = x * y
        return math.floor(area * self.const.f)
        """

        r = Rickle(y, load_lambda=True)

        area = r.get_area(x=10, y=10, z=10)

        self.assertEqual(area, 300)

        y = """
        const:
            f: 0.5
        functions:
            get_area:
              type: function
              name: get_area
              is_method: true
              args:
                 x: 10
                 y: 10
                 z: null
              import:
                 - math
              load: >
                 def get_area(self, x, y, z):
                    if not z is None:
                       area = (x * y) + (x * z) + (y * z)
                       area = 2 * area
                    else:
                       area = x * y
                    return math.floor(area * self.const.f)
                """

        r = Rickle(y, load_lambda=True)

        with self.assertRaises(AttributeError):
            area = r.functions.get_area(x=10, y=10, z=10)

    def test_hot_load_api(self):

        s = """
random_joke:
  type: api_json
  url: https://official-joke-api.appspot.com/random_joke
  expected_http_status: 200
  load_as_rick: true
  deep: true
  hot_load: true
        """

        r = Rickle(s)

        observed = r.random_joke()

        self.assertTrue(isinstance(observed, Rickle))

        d = r.dict()

        self.assertTrue(isinstance(d['random_joke'], dict))

        s = """
random_joke:
  type: api_json
  url: https://official-joke-api.appspot.com/random_joke
  expected_http_status: 200
  load_as_rick: true
  deep: true
        """

        r = Rickle(s)

        self.assertTrue(isinstance(r.random_joke, Rickle))


        d = r.dict()

        self.assertTrue(isinstance(d['random_joke'], dict))

    def test_hot_load_html(self):

        s = """
page:
    type: html_page
    url: https://zipfian.science
    expected_http_status: 200
    hot_load: true
        """

        r = Rickle(s)

        observed = r.page()

        self.assertTrue(isinstance(observed, str))

        d = r.dict()

        self.assertTrue(isinstance(d['page'], dict))

        s = """
page:
    type: html_page
    url: https://zipfian.science
    expected_http_status: 200
        """

        r = Rickle(s)

        self.assertTrue(isinstance(r.page, str))

        expected_meta = {'type': 'html_page',
                        'url': 'https://zipfian.science',
                        'headers': None,
                        'params': None,
                        'expected_http_status': 200,
                        'hot_load' : False
                    }

        actual_meta = r.meta('page')

        self.assertDictEqual(expected_meta, actual_meta)

        d = r.dict()

        self.assertTrue(isinstance(d['page'], str))

    def test_hot_load_file(self):
        s = """
another_rick:
   type: from_file
   file_path: './tests/placebos/test_config.json'
   hot_load: true
        """

        r = Rickle(s)

        observed = r.another_rick()

        self.assertTrue(isinstance(observed, str))

        d = r.dict()

        self.assertTrue(isinstance(d['another_rick'], dict))

        s = """
another_rick:
   type: from_file
   file_path: './tests/placebos/test_config.json'
        """

        r = Rickle(s)

        self.assertTrue(isinstance(r.another_rick, str))

        d = r.dict()

        self.assertTrue(isinstance(d['another_rick'], str))

    def test_serialise_list(self):

        s = """
BASICS:
    text: test
    dictionary:
        one: value
        two: value
    number: 2
    list:
        - one
        - two
        - four
        - name: John
          age: 20
        """

        r = Rickle(s, deep=True)

        self.assertTrue(isinstance(r.BASICS.list[-1], Rickle))

        d = r.dict()

        self.assertTrue(isinstance(d['BASICS']['list'][-1], dict))

    def test_strict(self):

        s = """
test:
    has: name
    get: name
        """

        with self.assertRaises(ValueError):
            r = BaseRickle(s, strict=True)

        with self.assertRaises(ValueError):
            r = Rickle(s, strict=True)

        r = BaseRickle(s, strict=False)
        with self.assertRaises(TypeError):
            r.test.has('get')

        r = Rickle(s, strict=False)
        with self.assertRaises(TypeError):
            r.test.has('get')

        r = BaseRickle()
        with self.assertRaises(ValueError):
            r.add_attr("__setattr__", True)

        r = Rickle()
        with self.assertRaises(ValueError):
            r.add_attr("__setattr__", True)

    def test_callable(self):
        test_yaml = """
root:
    func:
        type: function
        name: func
        args:
            x: 2
            y: 2
        load: >
            def func(x, y):
                return x ** y
        """

        r = Rickle(test_yaml, load_lambda=True)

        v = r('/root/func?x=2&y=3')

        self.assertEqual(v, 8)

        v = r('/root/func', x=3, y=2)

        self.assertEqual(v, 9)

        v = r('/root/func')

        self.assertEqual(v, 4)



import unittest
import os
from rickled.tools import Schema

class TestAdvanced(unittest.TestCase):

    def test_lists(self):
        schem_d = {
            'type': 'dict',
            'schema': {
                'values_empty': {
                    'type': 'list',
                    'schema': []
                },
                'values_filled': {
                    'type': 'list',
                    'length': -1,
                    'required': True,
                    'schema': [{'type': 'str'}]
                },
                'values_filled_max': {
                    'type': 'list',
                    'length': 1,
                    'schema': [{'type': 'str'}]
                },
                'values_any': {
                    'type': 'list',
                    'schema': [{'type': 'any'}]
                },
                'values_nullable': {
                    'type': 'list',
                    'nullable': True,
                    'schema': [{'type': 'bool'}]
                },
            }
        }

        test_all_good = {
            'values_empty': [1, 2, 3],
            'values_filled': ['a', 'b', 'c'],
            'values_filled_max': ['a']
        }

        test_missing_good = {
            'values_empty': [],
            'values_filled': ['a', 'b', 'c']
        }

        test_missing_fail = {
            'values_empty': [1, 2, 3],
            'values_filled_max': ['a']
        }

        test_type_err = {
            'values_empty': [],
            'values_filled': [1, 2, 3]
        }

        test_type_err_second = {
            'values_empty': [],
            'values_filled': ['a', 'b', 'c', 'd', 1, 'e']
        }

        test_too_long = {
            'values_empty': [1, 2, 3],
            'values_filled': ['a', 'b', 'c'],
            'values_filled_max': ['a', 'b']
        }

        test_mixed = {
            'values_empty': [],
            'values_filled': ['a', 'b', 'c'],
            'values_any': [8, '0.1', True, .99]
        }

        test_nullable_okay = {
            'values_empty': [],
            'values_filled': ['a', 'b', 'c'],
            'values_any': [8, '0.1', True, .99],
            'values_nullable': None
        }

        test_nullable_fail_type = {
            'values_empty': [],
            'values_filled': ['a', 'b', 'c'],
            'values_any': [8, '0.1', True, .99],
            'values_nullable': ['true', 'false']
        }

        test_nullable_fail_null = {
            'values_empty': [],
            'values_filled': ['a', 'b', 'c'],
            'values_any': None,
            'values_nullable': None
        }

        test_empty_null_fail = {
            'values_empty': None,
            'values_filled': ['a', 'b', 'c'],
            'values_filled_max': ['a']
        }

        test_empty_fail_type = {
            'values_empty': 'string',
            'values_filled': ['a', 'b', 'c'],
            'values_filled_max': ['a']
        }

        self.assertTrue(Schema.schema_validation(test_all_good, schem_d))

        self.assertTrue(Schema.schema_validation(test_missing_good, schem_d))

        self.assertFalse(Schema.schema_validation(test_missing_fail, schem_d))

        self.assertFalse(Schema.schema_validation(test_type_err, schem_d))
        self.assertFalse(Schema.schema_validation(test_type_err_second, schem_d))
        self.assertFalse(Schema.schema_validation(test_too_long, schem_d))

        self.assertTrue(Schema.schema_validation(test_mixed, schem_d))
        self.assertTrue(Schema.schema_validation(test_nullable_okay, schem_d))

        self.assertFalse(Schema.schema_validation(test_nullable_fail_type, schem_d))
        self.assertFalse(Schema.schema_validation(test_nullable_fail_null, schem_d))
        self.assertFalse(Schema.schema_validation(test_empty_fail_type, schem_d))
        self.assertFalse(Schema.schema_validation(test_empty_null_fail, schem_d))


    def test_primitives(self):
        schem_d = {
            'type': 'dict',
            'schema': {
                'A': {'type': 'int', 'required': True},
                'a': {'type': 'float'},
                'b': {'type': 'any'},
                'c': {'type': 'bool'},
                'd': {'type': 'str', 'nullable': True},
                'e': {'type': 'str', 'nullable': True, 'required': True}
            }
        }

        test_all_good = {
            'A': 10,
            'a': .90,
            'e': 'string'
        }

        test_missing_good = {
            'A': 10,
            'e': 'string'
        }

        test_missing_fail = {
            'b': 10,
            'c': .90,
            'e': 'string'
        }

        test_type_err = {
            'A': 10,
            'a': 90,
            'e': 'string'
        }

        test_type_any = {
            'A': 10,
            'a': .89,
            'b': .99,
            'e': 'string'
        }

        test_fail_null = {
            'A': None,
            'e': 'string'
        }

        test_fail_okay = {
            'A': 1,
            'd': None,
            'e': 'string'
        }

        test_fail_missing_nullable = {
            'A': 1,
            'd': None,
        }

        test_okay_nullable = {
            'A': 1,
            'd': None,
            'e': None
        }

        test_fail_nullable_type = {
            'A': 1,
            'd': None,
            'e': 1
        }

        self.assertTrue(Schema.schema_validation(test_all_good, schem_d))
        self.assertTrue(Schema.schema_validation(test_missing_good, schem_d))

        self.assertFalse(Schema.schema_validation(test_missing_fail, schem_d))
        self.assertFalse(Schema.schema_validation(test_type_err, schem_d))

        self.assertTrue(Schema.schema_validation(test_type_any, schem_d))

        self.assertFalse(Schema.schema_validation(test_fail_null, schem_d))

        self.assertTrue(Schema.schema_validation(test_fail_okay, schem_d))

        self.assertFalse(Schema.schema_validation(test_fail_missing_nullable, schem_d))
        self.assertTrue(Schema.schema_validation(test_okay_nullable, schem_d))

        self.assertFalse(Schema.schema_validation(test_fail_nullable_type, schem_d))



import unittest
from rickled import ObjectRickler, Rickle

class Testing:

    x = 0.2
    y= 'str'
    z = True
    b = 'asdasd'

    l = [1,2,3,4]
    l_a = [{ 'a' : 7}]

    d = {
        'p' : 0
    }

    def testing_function(self, a, b, c=1):
        return f'Hello {a}, {b}, and {c}, {self.l}'



class TestObjectRickler(unittest.TestCase):

    def test_to_rickle(self):
        t = Testing()

        rickler = ObjectRickler()

        d = rickler.deconstruct(t)

        self.assertTrue(True)

        rick = rickler.to_rickle(t, load_lambda=True)

        s = rick.testing_function(a='Jack', b='Sally', c='Justin?')

        self.assertTrue(True)

    def test_rickler_basic(self):

        t = Testing()

        rickler = ObjectRickler()

        d = rickler.deconstruct(t)

        d['testing_function_new'] = {'name' : 'testing_function_new',
                                     'type' : 'function',
                                     'is_method' : True,
                                     'args' : {'a' : None, 'b' : None, 'c' : 1},
                                     'load': """def testing_function_new(self, a, b, c=1):
    print(f\'{self.__class__.__name__} Howdy {a}, {b}, and {c}\')
"""}

        del d['testing_function']

        y = rickler.to_yaml_string(t)

        r = Rickle(d, deep=True, load_lambda=True)

        # r = rickler.to_rickle(t, deep=True, load_lambda=True)

        # r.testing_function_new('Milly', 'Willy')

        r.x = 99
        r.y = 99
        r.z = 99
        r.b = 99

        obj = rickler.from_rickle(r, Testing)

        obj.testing_function('Jack', 'Sally', 'Justin?')

        obj.testing_function_new('Jack', 'Sally')

        self.assertTrue(True)





# ASCCI Text found in this file was generated on
# https://patorjk.com/software/taag/
# Large heading using "Shaded Blocky" font
# Sub headings using "Small" font

import rickled.__version__ as ver
import argparse
from rickled.tools import cli_bcolors
from rickled.tools import Schema
from rickled.tools import Converter
from rickled.net import serve_rickle_http, serve_rickle_https
from rickled import Rickle
import re
import json
import yaml
import ast

def conv(args):
    Converter(input_files=args.i,
              output_files=args.o,
              input_directories=args.d,
              default_output_type=args.t,
              silent=args.s).do_convert()

def obj_get(args):
    import json
    import yaml
    if args:
        r = Rickle(args.i, load_lambda=args.l)
        v = r.get(args.key)
        if isinstance(v, Rickle):
            if args.t.lower() == 'json':
                print(v.to_json_string())
            else:
                print(v.to_yaml_string())
        elif isinstance(v, dict):
            if args.t.lower() == 'json':
                print(json.dumps(v))
            else:
                print(yaml.safe_dump(v))
        else:
            print(v)



def obj_set(args):
    if args:
        r = Rickle(args.i)
        r.set(args.key, args.value)
        if args.p:
            if args.t.lower() == 'json':
                print(r.to_json_string())
            else:
                print(r.to_yaml_string())
        if args.o:
            if args.t.lower() == 'json':
                r.to_json_file(args.o)
            else:
                r.to_yaml_file(args.o)


def obj_del(args):
    if args:
        r = Rickle(args.i)
        r.remove(args.key)
        if args.p:
            if args.t.lower() == 'json':
                print(r.to_json_string())
            else:
                print(r.to_yaml_string())
        if args.o:
            if args.t.lower() == 'json':
                r.to_json_file(args.o)
            else:
                r.to_yaml_file(args.o)

def obj_type(args):
    if args:
        r = Rickle(args.i, load_lambda=args.l)
        v = r.get(args.key)
        print(type(v))

def obj_search(args):
    if args:
        r = Rickle(args.i, load_lambda=args.l)
        paths = r.search_path(args.key)
        for p in paths:
            print(p)

def obj_func(args):
    def guess_parse(param_value):
        try:
            return ast.literal_eval(param_value)
        except Exception as exc:
            raise TypeError(f"Could not guess the parameter type, {exc}")
    def parse_type(type_name, param_value):
        type_name = type_name.strip().lower()
        if type_name == 'str':
            return str(param_value)
        elif type_name == 'int':
            try:
                return int(param_value)
            except:
                raise TypeError(f"Value '{param_value}' does not match type int")
        elif type_name == 'float':
            try:
                return float(param_value)
            except:
                raise TypeError(f"Value '{param_value}' does not match type int")
        elif type_name == 'bool':
            if param_value.strip().lower() == 'true':
                return True
            if param_value.strip().lower() == 'false':
                return False
            raise TypeError(f"Value '{param_value}' does not match type bool")
        elif type_name == 'list' or type_name == 'dict':
            try:
                return ast.literal_eval(param_value)
            except Exception as exc:
                raise TypeError(f"Could not guess the parameter type, {exc}")
        else:
            raise ValueError('Could not interpret type, only str, int, float, bool, list, dict accepted.')


    re_pat = re.compile(r"(.+?)=(.+)")
    if args:
        r = Rickle(args.i, load_lambda=args.l)

        params = dict()
        if args.params:
            for p in args.params:
                m = re_pat.match(p)
                param_name = m.group(1)
                param_value = m.group(2)
                if ':' in param_name:
                    param_name, ptype = param_name.split(':')
                    param_value = parse_type(ptype, param_value)
                elif args.x:
                    param_value = guess_parse(param_value)

                params[param_name] = param_value
        v = r(args.key, **params)

        if not v is None:
            if isinstance(v, Rickle):
                if args.t.lower() == 'json':
                    print(v.to_json_string())
                else:
                    print(v.to_yaml_string())
            elif isinstance(v, dict):
                if args.t.lower() == 'json':
                    print(json.dumps(v))
                else:
                    print(yaml.safe_dump(v))
            else:
                print(v)


def serve(args):
    rick = Rickle(args.f)

    if args.b:
        import webbrowser
        host = 'localhost' if args.a == '' else args.a
        scheme = 'https' if args.c and args.k else 'http'
        webbrowser.open(f'{scheme}://{host}:{args.p}', new=2)

    if args.c and args.k:
        serve_rickle_https(rickle=rick,
                           path_to_certificate=args.c,
                           path_to_private_key=args.k,
                           port=args.p,
                           interface=args.a
                           )
    else:
        serve_rickle_http(rickle=rick,
                          port=args.p,
                          interface=args.a
                          )

def check(args):
    Schema(input_files=args.i,
           input_directories=args.d,
           schema=args.c,
           output_dir=args.o,
           silent=args.s).do_validation()

def gen(args):
    Schema(input_files=args.i,
           input_directories=args.d,
           silent=args.s).do_generation()

def main():
    parser = argparse.ArgumentParser(
        prog='rickle',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=f"""
-----------------------------------------------------
{cli_bcolors.HEADER}YAML tools for Python{cli_bcolors.ENDC} (version {ver}).
""",
        epilog="-----------------------------------------------------"
    )
    parser.add_argument(
        "--version",
        "-v",
        help="show version number",
        action="version",
        version=f'%(prog)s {ver}'
    )

    subparsers = parser.add_subparsers()

    #################### CONV #####################
    # ░░      ░░░      ░░   ░░░  ░  ░░░░  ░
    # ▒  ▒▒▒▒  ▒  ▒▒▒▒  ▒    ▒▒  ▒  ▒▒▒▒  ▒
    # ▓  ▓▓▓▓▓▓▓  ▓▓▓▓  ▓  ▓  ▓  ▓▓  ▓▓  ▓▓
    # █  ████  █  ████  █  ██    ███    ███
    # ██      ███      ██  ███   ████  ████


    parser_conv = subparsers.add_parser('conv',
                                        help=f'Tool for {cli_bcolors.OKBLUE}converting{cli_bcolors.ENDC} files to or from YAML',
                                        description=f"""
---
{cli_bcolors.HEADER}Tool for converting files to or from YAML{cli_bcolors.ENDC}.
    """,
                                        epilog="-----------------------------------------------------"
                                        )

    parser_conv.add_argument('-i', type=str, help=f"{cli_bcolors.OKBLUE}input file{cli_bcolors.ENDC}(s) to convert",
                             nargs='+', metavar='input')
    parser_conv.add_argument('-d', type=str, help=f"{cli_bcolors.OKBLUE}directory{cli_bcolors.ENDC}(s) of input files",
                             default=None, nargs='+', metavar='dir')
    parser_conv.add_argument('-o', type=str, help=f"{cli_bcolors.OKBLUE}output file{cli_bcolors.ENDC} names",
                             nargs='+', metavar='output')
    parser_conv.add_argument('-t', type=str, help=f"default output {cli_bcolors.OKBLUE}file type{cli_bcolors.ENDC} (JSON, YAML)",
                             default='yaml', metavar='type')
    parser_conv.add_argument('-s', action='store_true', help=f"{cli_bcolors.OKBLUE}suppress{cli_bcolors.ENDC} verbose output", )

    parser_conv.set_defaults(func=conv)

    #################### OBJ #####################
    # ░░      ░░       ░░        ░
    # ▒  ▒▒▒▒  ▒  ▒▒▒▒  ▒▒▒▒▒▒▒  ▒
    # ▓  ▓▓▓▓  ▓       ▓▓▓▓▓▓▓▓  ▓
    # █  ████  █  ████  █  ████  █
    # ██      ██       ███      ██


    parser_obj = subparsers.add_parser('obj',
                                       help=f'Tool for {cli_bcolors.OKBLUE}accessing/manipulating{cli_bcolors.ENDC} YAML files',
                                       description=f"""
    ---
    {cli_bcolors.HEADER}Tool for accessing/manipulating YAML files{cli_bcolors.ENDC}.
        """,
                                       epilog="-----------------------------------------------------"
                                       )

    parser_obj.add_argument('-i', type=str, help=f"{cli_bcolors.OKBLUE}input file{cli_bcolors.ENDC} to read/modify",
                            metavar='input', required=True)
    parser_obj.add_argument('-o', type=str, help=f"{cli_bcolors.OKBLUE}output file{cli_bcolors.ENDC} to save modified",
                            metavar='output', required=False)
    parser_obj.add_argument('-t', type=str,
                            help=f"output {cli_bcolors.OKBLUE}type{cli_bcolors.ENDC} (JSON, YAML)",
                            default='yaml', metavar='type')
    parser_obj.add_argument('-p', action='store_true', help=f"{cli_bcolors.OKBLUE}print{cli_bcolors.ENDC} output",
                            default=False)
    parser_obj.add_argument('-l', action='store_true', help=f"load {cli_bcolors.OKBLUE}lambda{cli_bcolors.ENDC} types",
                            default=False)

    subparsers_obj = parser_obj.add_subparsers()

    #################### OBJ - GET #####################
    #   ___ ___ _____
    #  / __| __|_   _|
    # | (_ | _|  | |
    #  \___|___| |_|

    get_obj_parser = subparsers_obj.add_parser('get',
                                               help=f'Tool for {cli_bcolors.OKBLUE}getting{cli_bcolors.ENDC} values from YAML files',
                                               description=f"""
        ---
        {cli_bcolors.HEADER}Tool for getting values from YAML files{cli_bcolors.ENDC}.
            """,
                                               epilog="-----------------------------------------------------"
                                               )

    get_obj_parser.add_argument('key', type=str, help=f"{cli_bcolors.OKBLUE}Key{cli_bcolors.ENDC} to get value",
                                metavar='key')

    get_obj_parser.set_defaults(func=obj_get)

    #################### OBJ - SET #####################
    #  ___ ___ _____
    # / __| __|_   _|
    # \__ \ _|  | |
    # |___/___| |_|

    set_obj_parser = subparsers_obj.add_parser('set',
                                               help=f'Tool for {cli_bcolors.OKBLUE}setting{cli_bcolors.ENDC} values in YAML files',
                                               description=f"""
            ---
            {cli_bcolors.HEADER}Tool for setting values in YAML files{cli_bcolors.ENDC}.
                """,
                                               epilog="-----------------------------------------------------"
                                               )

    set_obj_parser.add_argument('key', type=str, help=f"{cli_bcolors.OKBLUE}Key{cli_bcolors.ENDC} to set value",
                                metavar='key')
    set_obj_parser.add_argument('value', type=str, help=f"{cli_bcolors.OKBLUE}Value{cli_bcolors.ENDC} to set",
                                metavar='value')

    set_obj_parser.set_defaults(func=obj_set)

    #################### OBJ - DEL #####################
    #  ___  ___ _
    # |   \| __| |
    # | |) | _|| |__
    # |___/|___|____|

    del_obj_parser = subparsers_obj.add_parser('del',
                                               help=f'Tool for {cli_bcolors.OKBLUE}deleting{cli_bcolors.ENDC} keys in YAML files',
                                               description=f"""
                ---
                {cli_bcolors.HEADER}Tool for deleting keys in YAML files{cli_bcolors.ENDC}.
                    """,
                                               epilog="-----------------------------------------------------"
                                               )

    del_obj_parser.add_argument('key', type=str, help=f"{cli_bcolors.OKBLUE}Key{cli_bcolors.ENDC} to delete",
                                metavar='key')

    del_obj_parser.set_defaults(func=obj_del)

    #################### OBJ - TYPE #####################
    #  _______   _____ ___
    # |_   _\ \ / / _ \ __|
    #   | |  \ V /|  _/ _|
    #   |_|   |_| |_| |___|

    type_obj_parser = subparsers_obj.add_parser('type',
                                                help=f'Tool for {cli_bcolors.OKBLUE}checking type{cli_bcolors.ENDC} of keys in YAML files',
                                                description=f"""
                    ---
                    {cli_bcolors.HEADER}Tool for checking type of keys in YAML files{cli_bcolors.ENDC}.
                        """,
                                                epilog="-----------------------------------------------------"
                                                )

    type_obj_parser.add_argument('key', type=str, help=f"{cli_bcolors.OKBLUE}Key{cli_bcolors.ENDC} to check",
                                 metavar='key')

    type_obj_parser.set_defaults(func=obj_type)

    #################### OBJ - SEARCH #####################
    #  ___ ___   _   ___  ___ _  _
    # / __| __| /_\ | _ \/ __| || |
    # \__ \ _| / _ \|   / (__| __ |
    # |___/___/_/ \_\_|_\\___|_||_|

    search_obj_parser = subparsers_obj.add_parser('search',
                                                  help=f'Tool for {cli_bcolors.OKBLUE}searching{cli_bcolors.ENDC} keys in YAML files',
                                                  description=f"""
                    ---
                    {cli_bcolors.HEADER}Tool for searching keys in YAML files{cli_bcolors.ENDC}.
                        """,
                                                  epilog="-----------------------------------------------------"
                                                  )

    search_obj_parser.add_argument('key', type=str, help=f"{cli_bcolors.OKBLUE}Key{cli_bcolors.ENDC} to search",
                                   metavar='key')

    search_obj_parser.set_defaults(func=obj_search)

    #################### OBJ - FUNC #####################
    #  ___ _   _ _  _  ___
    # | __| | | | \| |/ __|
    # | _|| |_| | .` | (__
    # |_|  \___/|_|\_|\___|

    func_obj_parser = subparsers_obj.add_parser('func',
                                                help=f'Tool for {cli_bcolors.OKBLUE}executing function{cli_bcolors.ENDC} defined in YAML files',
                                                description=f"""
                    ---
                    {cli_bcolors.HEADER}Tool for executing function defined in YAML files{cli_bcolors.ENDC}.
                        """,
                                                epilog="-----------------------------------------------------"
                                                )

    func_obj_parser.add_argument('-x', action='store_true', help=f"{cli_bcolors.OKBLUE}infer parameter{cli_bcolors.ENDC} types",
                                 default=False)
    func_obj_parser.add_argument('key', type=str, help=f"{cli_bcolors.OKBLUE}Key{cli_bcolors.ENDC} (name) of function",
                                 metavar='key')
    func_obj_parser.add_argument('params', type=str, help=f"{cli_bcolors.OKBLUE}Params{cli_bcolors.ENDC} for function",
                                 metavar='params', nargs='+')

    func_obj_parser.set_defaults(func=obj_func)


    #################### SERVE #####################
    # ░░      ░░        ░       ░░  ░░░░  ░        ░
    # ▒  ▒▒▒▒▒▒▒  ▒▒▒▒▒▒▒  ▒▒▒▒  ▒  ▒▒▒▒  ▒  ▒▒▒▒▒▒▒
    # ▓▓      ▓▓      ▓▓▓       ▓▓▓  ▓▓  ▓▓      ▓▓▓
    # ███████  █  ███████  ███  ████    ███  ███████
    # ██      ██        █  ████  ████  ████        █


    parser_serve = subparsers.add_parser('serve',
                                         help=f'Tool for serving YAML via {cli_bcolors.OKBLUE}http(s){cli_bcolors.ENDC}',
                                         description=f"""
    ---
    {cli_bcolors.HEADER}Tool for serving YAML via http(s){cli_bcolors.ENDC}.
        """,
                                         epilog="-----------------------------------------------------"
                                         )

    parser_serve.add_argument('-f', type=str,
                              help=f"{cli_bcolors.OKBLUE}YAML{cli_bcolors.ENDC} or {cli_bcolors.OKBLUE}JSON{cli_bcolors.ENDC} file to serve",
                              metavar='file')
    # TODO implement config
    # parser_serve.add_argument('-c', type=str, help=f"{cli_bcolors.OKBLUE}config{cli_bcolors.ENDC} file path",
    #                           default=None, metavar='config')
    parser_serve.add_argument('-a', type=str, help=f"{cli_bcolors.OKBLUE}host address{cli_bcolors.ENDC}",
                              default='', metavar='address')
    parser_serve.add_argument('-p', type=int, help=f"{cli_bcolors.OKBLUE}port{cli_bcolors.ENDC} number",
                              default=8080, metavar='port')
    parser_serve.add_argument('-k', type=str, help=f"{cli_bcolors.OKBLUE}private key{cli_bcolors.ENDC} file path",
                              default=None, metavar='privkey')
    parser_serve.add_argument('-c', type=str, help=f"{cli_bcolors.OKBLUE}SSL certificate{cli_bcolors.ENDC} file path",
                              default=None, metavar='cert')
    parser_serve.add_argument('-b', action='store_true', help=f"open URL in {cli_bcolors.OKBLUE}browser{cli_bcolors.ENDC}", )
    parser_serve.set_defaults(func=serve)

    #################### SCHEMA #####################
    # ░░      ░░░      ░░  ░░░░  ░        ░  ░░░░  ░░      ░░
    # ▒  ▒▒▒▒▒▒▒  ▒▒▒▒  ▒  ▒▒▒▒  ▒  ▒▒▒▒▒▒▒   ▒▒   ▒  ▒▒▒▒  ▒
    # ▓▓      ▓▓  ▓▓▓▓▓▓▓        ▓      ▓▓▓        ▓  ▓▓▓▓  ▓
    # ███████  █  ████  █  ████  █  ███████  █  █  █        █
    # ██      ███      ██  ████  █        █  ████  █  ████  █


    parser_schema = subparsers.add_parser('schema',
                                          help=f'Tool for generating and checking {cli_bcolors.OKBLUE}schemas{cli_bcolors.ENDC} of YAML files',
                                          description=f"""
    ---
    {cli_bcolors.HEADER}Tool for generating and checking schemas of YAML files{cli_bcolors.ENDC}.
        """,
                                          epilog="-----------------------------------------------------"
                                          )

    schema_subparsers = parser_schema.add_subparsers()

    #################### SCHEMA - CHECK ############
    #   ___ _  _ ___ ___ _  __
    #  / __| || | __/ __| |/ /
    # | (__| __ | _| (__| ' <
    #  \___|_||_|___\___|_|\_\

    parser_schema_check = schema_subparsers.add_parser('check',
                                                       help=f'Tool for {cli_bcolors.OKBLUE}checking{cli_bcolors.ENDC} schemas of YAML files',
                                                       description=f"""
        ---
        {cli_bcolors.HEADER}Tool for checking schemas of YAML files{cli_bcolors.ENDC}.
            """,
                                                       epilog="-----------------------------------------------------"
                                                       )


    parser_schema_check.add_argument('-i', type=str, help=f"{cli_bcolors.OKBLUE}input file{cli_bcolors.ENDC}(s) to check",
                                     nargs='+', metavar='input')
    parser_schema_check.add_argument('-d', type=str, help=f"{cli_bcolors.OKBLUE}directory{cli_bcolors.ENDC}(s) of files to check",
                                     default=None, nargs='+', metavar='dir')

    parser_schema_check.add_argument('-c', type=str, help=f"{cli_bcolors.OKBLUE}schema definition file{cli_bcolors.ENDC} to compare",
                                     metavar='schema')

    parser_schema_check.add_argument('-o', type=str,
                                     help=f"{cli_bcolors.OKBLUE}output directory{cli_bcolors.ENDC} to move failed files to",
                                     metavar='output')

    parser_schema_check.add_argument('-s', action='store_true', help=f"{cli_bcolors.OKBLUE}suppress{cli_bcolors.ENDC} verbose output", )

    parser_schema_check.set_defaults(func=check)

    #################### SCHEMA - GEN ############
    #   ___ ___ _  _
    #  / __| __| \| |
    # | (_ | _|| .` |
    #  \___|___|_|\_|

    parser_schema_gen = schema_subparsers.add_parser('gen',
                                                     help=f'Tool for {cli_bcolors.OKBLUE}generating{cli_bcolors.ENDC} schemas of YAML files',
                                                     description=f"""
            ---
            {cli_bcolors.HEADER}Tool for generating schemas of YAML files{cli_bcolors.ENDC}.
                """,
                                                     epilog="-----------------------------------------------------"
                                                     )

    parser_schema_gen.add_argument('-i', type=str, help=f"{cli_bcolors.OKBLUE}input file{cli_bcolors.ENDC}(s) to generate from",
                                   nargs='+', metavar='input')
    parser_schema_gen.add_argument('-d', type=str,
                                   help=f"{cli_bcolors.OKBLUE}directory{cli_bcolors.ENDC}(s) of files to generate from",
                                   default=None, nargs='+', metavar='dir')

    parser_schema_gen.add_argument('-s', action='store_true',
                                   help=f"{cli_bcolors.OKBLUE}suppress{cli_bcolors.ENDC} verbose output", )

    parser_schema_gen.set_defaults(func=gen)

    #################################################

    args = parser.parse_args()

    args.func(args)

if __name__ == "__main__":
    main()


from typing import List, Union
from pathlib import Path
import yaml
import json
import os

class cli_bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class Schema:
    """
    Tool for inferring schemas from Python dicts and validating schemas.

    Args:
        input_files (list): List of input file paths (default = None).
        input_directories (list): List of directories (default = None).
        schema (str, dict): The dict definition of schema or path to schema file (default = None).
        output_dir (str): Directory to move output files to (default = None).
        silent (bool): Suppress verbose output (default = None).
    """

    def __init__(self,
                 input_files: List[str] = None,
                 input_directories: List[str] = None,
                 schema: Union[str, dict] = None,
                 output_dir: str = None,
                 silent: bool = False,
                 ):
        self.input_files = input_files
        self.input_directories = input_directories

        if isinstance(schema, str):
            self.schema = Converter.infer_read_file_type(schema)
        else:
            self.schema = schema

        self.output_dir = output_dir

        self.silent = silent

    def do_generation(self):
        """
        Generates schema files for the given input files.
        """
        if self.input_files is None and self.input_directories is None:
            raise ValueError("Either input_files or input_directories must be defined!")

        if self.input_files is None and not self.input_directories is None:
            self.input_files = list()
            for d in self.input_directories:
                dir_path = Path(d)
                self.input_files.extend(list(dir_path.glob("*.yaml")))
                self.input_files.extend(list(dir_path.glob("*.json")))


        self.output_files = list()
        for input_file in self.input_files:
            self.output_files.append(f"{os.path.splitext(input_file)[0]}.schema.yaml")


        zipped = zip(self.input_files, self.output_files)

        for pair in zipped:
            try:
                input_data = Converter.infer_read_file_type(pair[0])
                output_file = Path(pair[1])

                schema = Schema.generate_schema_from_obj(input_data)

                with output_file.open("w") as fout:
                    yaml.safe_dump(schema, fout)

                if not self.silent:
                    print(f"{cli_bcolors.OKBLUE}{pair[0]}{cli_bcolors.ENDC} -> {cli_bcolors.OKBLUE}{pair[1]}{cli_bcolors.ENDC}")
                continue
            except Exception as exc:
                if not self.silent:
                    print(f"{cli_bcolors.FAIL}{str(exc)}{cli_bcolors.ENDC}")
                continue
    def do_validation(self):
        """
        Validates input files against schema definition.

        Returns:
            list: List of files that did not pass validation.
        """
        failed_validation = list()

        if self.input_files is None and self.input_directories is None:
            raise ValueError("Either input_files or input_directories must be defined!")

        if self.input_files is None and not self.input_directories is None:
            self.input_files = list()
            for d in self.input_directories:
                dir_path = Path(d)
                self.input_files.extend(list(dir_path.glob("*.yaml")))
                self.input_files.extend(list(dir_path.glob("*.json")))

        for file in self.input_files:
            try:
                input_data = Converter.infer_read_file_type(file)

                passed = Schema.schema_validation(input_data, self.schema, no_print=self.silent)

                if not passed:
                    failed_validation.append(file)

                if not self.silent:
                    result = f"{cli_bcolors.OKGREEN}OK{cli_bcolors.ENDC}" if passed else f"{cli_bcolors.FAIL}FAIL{cli_bcolors.ENDC}"
                    print(f"{cli_bcolors.OKBLUE}{file}{cli_bcolors.ENDC} -> {result}")
                continue
            except Exception as exc:
                if not self.silent:
                    print(f"{cli_bcolors.FAIL}{str(exc)}{cli_bcolors.ENDC}")
                continue

        if self.output_dir:
            outdir = Path(self.output_dir)
            if not outdir.is_dir():
                outdir.mkdir()

            for f in failed_validation:
                f_path = Path(f)
                f_path.rename(outdir / f_path.name)


        return failed_validation

    @staticmethod
    def _extract_data_types(value: Union[list, dict, str, int, float, bool]):
        if isinstance(value, list):
            schema = list()
            for v in value:
                schema.append(Schema._extract_data_types(v))
            return schema
        if isinstance(value, dict):
            schema = dict()
            for k, v in value.items():
                schema[k] = Schema._extract_data_types(v)
            return schema
        if isinstance(value, str):
            return str
        if isinstance(value, int):
            return int
        if isinstance(value, float):
            return float
        if isinstance(value, bool):
            return bool

    @staticmethod
    def _data_types_to_schema(value: Union[list, dict, str, int, float, bool, None]):

        if value == str:
            return {'type': 'str'}
        if value == int:
            return {'type': 'int'}
        if value == float:
            return {'type': 'float'}
        if value == bool:
            return {'type': 'bool'}
        if value is None:
            return {'type': 'any'}

        if isinstance(value, dict):
            named_schema = {'type': 'dict', 'schema': dict()}
            for k, v in value.items():
                named_schema['schema'][k] = Schema._data_types_to_schema(v)
            return named_schema

        if isinstance(value, list):
            named_schema = {'type': 'list', 'schema': list()}
            list_data_types = set(value)

            if len(list_data_types) > 1:
                named_schema['schema'].append(Schema._data_types_to_schema(None))
            if len(list_data_types) == 1:
                named_schema['schema'].append(Schema._data_types_to_schema(list_data_types.pop()))
            return named_schema

    @staticmethod
    def generate_schema_from_obj(obj):
        """
        Generate a schema definition from a Python object.

        Args:
            obj: Dict like object.

        Returns:
            dict: Schema detected from obj.

        """
        rep = Schema._extract_data_types(obj)
        return Schema._data_types_to_schema(rep)

    @staticmethod
    def schema_validation(obj, schema: dict, path: str = 'root', no_print: bool = False) -> bool:
        """
        Validates if obj conforms to schema.

        Args:
            obj: The object to check.
            schema (dict): The schema in dict form.
            path (str): The current path of the object tree (default = 'root').
            no_print (bool): If failures should not be printed (default = False).

        Returns:
            bool: True if the object conforms.
        """
        new_path = path
        if schema['type'] == 'dict':
            for k, v in schema['schema'].items():
                new_path = f"{new_path}->{k}"
                req = v.get('required', False)
                nullable = v.get('nullable', False)
                present = k in obj.keys()
                if not present:
                    if req:
                        if not no_print:
                            print(f"Required {cli_bcolors.FAIL}'{k}'{cli_bcolors.ENDC} (per schema {v}),\n In {obj},\n Path {cli_bcolors.WARNING}{new_path}{cli_bcolors.ENDC}")
                        return False
                    else:
                        new_path = path
                        continue

                null_no_type = nullable & ((v['type'] == 'any') | (type(obj[k]).__name__ == 'NoneType'))
                null_with_type = nullable & (v['type'] != 'any') & (type(obj[k]).__name__ == v['type'])
                non_null = (nullable is False) & ((v['type'] == 'any') | (type(obj[k]).__name__ == v['type']))

                if not (null_no_type or null_with_type or non_null):
                    if not no_print:
                        print(
                        f"Type '{k}' == {cli_bcolors.FAIL}'{type(obj[k]).__name__}'{cli_bcolors.ENDC},\n Required type {cli_bcolors.OKBLUE}'{v['type']}'{cli_bcolors.ENDC} (per schema {v}),\n In {obj},\n Path {cli_bcolors.WARNING}{new_path}{cli_bcolors.ENDC}")
                    return False

                if v['type'] in ['dict', 'list']:
                    if not Schema.schema_validation(obj[k], v, path=new_path, no_print=no_print):
                        return False
                new_path = path
            return True
        if schema['type'] == 'list':

            nullable = schema.get('nullable', False)

            if nullable and obj is None:
                return True

            length = schema.get('length', -1)
            obj_length = len(obj)
            schema_len = len(schema['schema'])

            if length > -1 and obj_length != length:
                if not no_print:
                    print(
                    f"Length '{obj}' == {cli_bcolors.FAIL}{obj_length}{cli_bcolors.ENDC},\n Required length {cli_bcolors.OKBLUE}{length}{cli_bcolors.ENDC} (per schema {schema['schema']}),\n In {obj},\n Path {cli_bcolors.WARNING}{new_path}{cli_bcolors.ENDC}")
                return False

            if schema_len > 0:
                single_type = schema['schema'][0]

                for i in range(obj_length):
                    new_path = f"{new_path}->[{i}]"
                    o = obj[i]
                    if single_type['type'] != 'any' and type(o).__name__ != single_type['type']:
                        if not no_print:
                            print(
                            f"Type '{o}' == {cli_bcolors.FAIL}'{type(o).__name__}'{cli_bcolors.ENDC},\n Required type {cli_bcolors.OKBLUE}'{single_type['type']}'{cli_bcolors.ENDC} (per schema {single_type}),\n In {o},\n Path {cli_bcolors.WARNING}{new_path}{cli_bcolors.ENDC}")
                        return False
                    if single_type['type'] in ['dict', 'list']:
                        if not Schema.schema_validation(o, single_type, path=new_path, no_print=no_print):
                            return False
                    new_path = path
            return True

class Converter:
    """
    Tool for converting between YAML and JSON, mainly to be used in CLI.

    Args:
        input_files (list): List of input file paths (default = None).
        input_directories (list): List of directories (default = None).
        output_files (list): List of output file paths (default = None).
        default_output_type (str): Default type to convert to if output files are not provided (default = 'yaml').
        silent (bool): Suppress verbose output (default = None).
    """

    def __init__(self,
                 input_files: List[str] = None,
                 input_directories: List[str] = None,
                 output_files: List[str] = None,
                 default_output_type: str = 'yaml',
                 silent: bool = False,
                 ):
        self.input_files = input_files
        self.input_directories = input_directories
        self.output_files = output_files
        self.default_output_type = default_output_type
        self.silent = silent

    @staticmethod
    def infer_read_file_type(file_path: str):
        """
        Infer the file type and return loaded contents. By default, the type is inferred from the suffix of the
        file path. Thus, `file.yaml` will be read as a YAML file. If the file extension is not known, the file will be
        read and tried to be loaded as both types.

        Raises:
             ValueError: If the type could not be inferred.

        Args:
            file_path (str): Input file path to read.

        Returns:
            dict: Loaded YAML (or JSON).
        """
        input_file = Path(file_path)

        suffix = input_file.suffix.lower() if input_file.suffix else None

        if suffix == '.json':
            with input_file.open("r") as fin:
                return json.load(fin)

        if suffix in ['.yaml', '.yml']:
            with input_file.open("r") as fin:
                return yaml.safe_load(fin)

        try:
            with input_file.open("r") as fin:
                return json.load(fin)
        except:
            pass

        try:
            with input_file.open("r") as fin:
                return yaml.safe_load(fin)
        except:
            raise ValueError(f"Input file {input_file.name} could not be inferred")

    def do_convert(self):
        """
        Converts all input files to output type(s).
        """
        if self.input_files is None and self.input_directories is None:
            raise ValueError("Either input_files or input_directories must be defined!")

        if self.input_files is None and not self.input_directories is None:
            # set output to none as it should not be defined in this scenario
            self.output_files = None
            self.input_files = list()
            for d in self.input_directories:
                dir_path = Path(d)
                self.input_files.extend(list(dir_path.glob("*.yaml")))
                self.input_files.extend(list(dir_path.glob("*.json")))


        if self.output_files is None:
            self.output_files = list()
            for input_file in self.input_files:
                self.output_files.append(f"{os.path.splitext(input_file)[0]}.{self.default_output_type}")


        zipped = zip(self.input_files, self.output_files)

        for pair in zipped:
            try:
                input_data = Converter.infer_read_file_type(pair[0])
                output_file = Path(pair[1])

                suffix = output_file.suffix.lower() if output_file.suffix else f".{self.default_output_type}"

                if suffix == '.yaml':
                    with output_file.open("w") as fout:
                        yaml.safe_dump(input_data, fout)

                if suffix == '.json':
                    with output_file.open("w") as fout:
                        json.dump(input_data, fout)

                if not self.silent:
                    print(f"{cli_bcolors.OKBLUE}{pair[0]}{cli_bcolors.ENDC} -> {cli_bcolors.OKBLUE}{pair[1]}{cli_bcolors.ENDC}")
                continue
            except Exception as exc:
                if not self.silent:
                    print(f"{cli_bcolors.FAIL}{str(exc)}{cli_bcolors.ENDC}")
                continue


import sys
import traceback
import warnings
import json

from rickled import Rickle

try:
    from twisted.web import server, resource
    from twisted.internet import reactor, endpoints, ssl
    from twisted.python import log
except ModuleNotFoundError as exc:
    warnings.warn('Required Python package not found.', ImportWarning)
except ImportError as exc:
    warnings.warn('Required Python package "twisted" not found.', ImportWarning)

try:
    from twisted.internet import ssl
except ModuleNotFoundError as exc:
    warnings.warn('Required Python package not found.', ImportWarning)
except ImportError as exc:
    warnings.warn('Required Python package "pyopenssl" not found.', ImportWarning)


class HttpResource(resource.Resource):
    isLeaf = True
    numberRequests = 0

    def __init__(self, rickle):
        self.rickle = rickle
        super().__init__()

    def render_GET(self, request):

        uri = request.uri.decode("utf-8")

        content = self.rickle(uri)

        try:
            if isinstance(content, Rickle):
                request.setHeader(b"content-type", b"application/json")
                response = content.dict(True)
                response = json.dumps(response)
            elif isinstance(content, dict) or isinstance(content, list):
                request.setHeader(b"content-type", b"application/json")
                response = json.dumps(content)
            elif isinstance(content, str):
                request.setHeader(b"content-type", b"text/html")
                response = content
            else:
                response = content
        except:
            status_code = 500
            request.setResponseCode(status_code)
            request.setHeader(b"content-type", b"text/html")
            response = traceback.format_exc()


        return response.encode("utf-8")

def serve_rickle_http(rickle, port: int = 8080, interface: str = ''):
    log.startLogging(sys.stdout)
    site = server.Site(HttpResource(rickle))
    reactor.listenTCP(port, site, interface=interface)
    reactor.run()

def serve_rickle_https(rickle, path_to_private_key: str, path_to_certificate: str, port: int = 8080, interface: str = ''):
    log.startLogging(sys.stdout)
    ssl_context = ssl.DefaultOpenSSLContextFactory(
        path_to_private_key,
        path_to_certificate,
    )

    site = server.Site(HttpResource(rickle))

    reactor.listenSSL(port, site, ssl_context, interface=interface)
    reactor.run()



from setuptools import setup, find_packages
import os
from datetime import datetime

here = os.path.abspath(os.path.dirname(__file__))

about = {}

with open(os.path.join(here, "rickled", "__version__.py")) as f:
    exec(f.read(), about)

# Get the long description from the README file
long_description_file = os.path.join(here, 'pip_description.md')
if os.path.isfile(long_description_file):
    with open(long_description_file, "r", encoding='utf-8') as f:
        long_description = f.read()
        long_description = long_description.format(pypi_metdata_release_date=datetime.today().strftime('%Y-%m-%d'),
                                                   pypi_metdata_version_number=about["__version__"])
else:
    long_description = "It's Rickle Pick!"

if os.path.isfile(os.path.join(here, 'requirements.txt')):
    with open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as f:
        pipreq = f.readlines()
        # remove pip flag
        if '-i http' in pipreq[0]:
            pipreq.pop(0)
else:
    pipreq = ['pyyaml', 'requests']

setup(
    name="rickled",
    version=about["__version__"],
    description='Tools for pickling Python objects in a completely different way',
    long_description_content_type='text/markdown',
    long_description=long_description,
    license='Apache 2.0',
    keywords = ['Pickle', 'Python', 'YAML', 'JSON'],
    author='Zipfian Science',
    author_email='about@zipfian.science',
    zip_safe=False,
    url='https://github.com/Zipfian-Science/rickle',
    download_url='https://github.com/Zipfian-Science/rickle/archive/v_01.tar.gz',
    packages=find_packages(".", exclude=("tests", "dist", "deploy", "egg-info")),
    include_package_data=True,
    install_requires=pipreq,
    package_dir={'.': 'rickled'},
    extras_require={
        'twisted':  ['twisted'],
        'pyopenssl':  ['pyopenssl']
    },
    entry_points={
        'console_scripts': [
            'rickle = rickled.cli:main',
        ],
    },
    classifiers=[
            'Intended Audience :: Developers',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Topic :: Software Development']
)



import unittest
import os
from rickled import BaseRickle, Rickle

class TestPickles(unittest.TestCase):

    def test_base_config(self):
        # Test normal dict
        test_dict = {'A' : 1, 'l' : [1, { 'deep' : 'hole'}], 'B' : { 'k' : 'v'}}
        test_conf = BaseRickle(test_dict)

        self.assertEquals(test_conf.A, 1)
        self.assertListEqual(test_conf.l, [1,{ 'deep' : 'hole'}])
        self.assertEquals(test_conf.B.k, 'v')

        # Test deep dict
        test_conf = BaseRickle(test_dict, deep=True)

        self.assertEquals(test_conf.l[-1].deep, 'hole')

        # Test YAML and JSON doc loading with file path

        test_conf_yaml = BaseRickle('./tests/placebos/test_config.yaml', deep=True)
        test_conf_json = BaseRickle('./tests/placebos/test_config.json', deep=True)

        self.assertGreater(len(test_conf_yaml), 0)
        self.assertGreater(len(test_conf_json), 0)

        # Test YAML and JSON doc loading with file stream

        with open('./tests/placebos/test_config.yaml', 'r') as fs_y, open('./tests/placebos/test_config.json', 'r') as fs_j:
            test_conf_yaml = BaseRickle(fs_y, deep=True)
            test_conf_json = BaseRickle(fs_j, deep=True)

            self.assertGreater(len(test_conf_yaml), 0)
            self.assertGreater(len(test_conf_json), 0)

        # Test string loading
        yaml_string = """
                ONE: "value"
                """
        json_string = """
                { "ONE" : "value"}
                """
        test_conf_yaml = BaseRickle(yaml_string, deep=True)
        test_conf_json = BaseRickle(json_string, deep=True)

        self.assertTrue(test_conf_yaml == test_conf_json)
        self.assertEquals(test_conf_yaml.ONE, "value")
        self.assertEquals(test_conf_json.ONE, "value")

    def test_base_set(self):
        test = {
            'level_one' : {
                'level_two' : {
                    'level_three' : 1
                }
            }
        }

        rickle = BaseRickle(test)

        rickle.set('/level_one/level_two/level_three', 99)

        with self.assertRaises(KeyError):
            rickle.set('level_one/level_two/level_three', -1)

        with self.assertRaises(NameError):
            rickle.set('/', -1)

        self.assertEquals(rickle.level_one.level_two.level_three, 99)

        rickle.set('level_one', True)

        self.assertTrue(rickle.level_one)

        rickle['planet'] = 'Hello Uranus!'

        self.assertEquals(rickle.planet, 'Hello Uranus!')

        rickle['planet'] = 'Hello Venus :)'

        self.assertEquals(rickle.planet, 'Hello Venus :)')

    def test_base_config_add_attr(self):
        test_conf = BaseRickle()

        test_conf.add_attr('new_var', 42)

        self.assertEquals(test_conf.new_var, 42)

        test_conf.new_var += 1

        self.assertEquals(test_conf.new_var, 43)

        expected_meta = {'type': 'attr', 'value': 42}

        actual_meta = test_conf.meta('new_var')

        self.assertDictEqual(expected_meta, actual_meta)

    def test_extended_config(self):
        # Test normal dict
        test_dict = {'A' : 1, 'l' : [1, { 'deep' : 'hole'}], 'B' : { 'k' : 'v'}}
        test_conf = Rickle(test_dict)

        self.assertEquals(test_conf.A, 1)
        self.assertListEqual(test_conf.l, [1,{ 'deep' : 'hole'}])
        self.assertEquals(test_conf.B.k, 'v')

        # Test deep dict
        test_conf = Rickle(test_dict, deep=True)

        self.assertEquals(test_conf.l[-1].deep, 'hole')

        # Test extended operations (OS ENV, functions)
        test_dict = {'user' : {
                        'type'  : 'env',
                        'load': 'USERNAME'
                    },
                     'func' : {
                        'type'  : 'lambda',
                        'load': 'lambda x: x+1'
                     }
        }

        test_conf = Rickle(test_dict, load_lambda=True)

        expected_username = os.getenv('USERNAME')

        self.assertEquals(test_conf.user, expected_username)
        self.assertEquals(test_conf.func(41), 42)

    def test_config_get_search(self):
        test_conf_yaml = BaseRickle('./tests/placebos/test_config.yaml', deep=True)

        value = test_conf_yaml.get('one')

        self.assertIsNone(value)

        value = test_conf_yaml.get('one', 'VALUE')

        self.assertEquals(value, 'VALUE')

        value = test_conf_yaml.get('one', do_recursive=True)

        self.assertEquals(value, 'value')

        value = test_conf_yaml.get('ONE')

        self.assertIsNone(value)

        value = test_conf_yaml.get('ONE', 'default_one', do_recursive=True)

        self.assertEquals(value, 'default_one')

        value = test_conf_yaml.get('USERNAME', do_recursive=True)

        self.assertIsInstance(value, BaseRickle)
        self.assertEquals(value.type, 'env')


        ###########
        # Test path like get

        value = test_conf_yaml.get('/BASICS/dictionary/one')

        self.assertEquals(value, 'value')

        ###########
        # Test dict like accesss

        v = test_conf_yaml['MYSQL']

        self.assertDictEqual(v.dict(), {'conn' : '127.0.0.1', 'usr' : 'test'})

        with self.assertRaises(KeyError):
            v = test_conf_yaml['asdasdasdasd']

    def test_config_remove(self):
        test_conf_yaml = BaseRickle('./tests/placebos/test_config.yaml', deep=True)

        test_conf_yaml.remove('/MYSQL/conn')

        self.assertDictEqual(test_conf_yaml.MYSQL.dict(), {'usr' : 'test'})

        test_conf_yaml = BaseRickle('./tests/placebos/test_config.yaml', deep=True)

        del test_conf_yaml['MYSQL']

        with self.assertRaises(KeyError):
            v = test_conf_yaml['MYSQL']

    def test_config_to_dict(self):
        test_dict = {'user': {
            'type': 'env',
            'load': 'USERNAME'
            },
            'func': {
                'type': 'lambda',
                'load': 'lambda x: x+1'
            }
        }

        test_conf = BaseRickle(test_dict)

        self.assertDictEqual(test_conf.dict(), test_dict)

    def test_config_keys(self):
        test_dict = {'user': {
            'type': 'env',
            'load': 'USERNAME'
            },
            'func': {
                'type': 'lambda',
                'load': 'lambda x: x+1'
            }
        }

        test_conf = Rickle(test_dict)

        for k in test_conf.keys():
            self.assertIn(k, ['user', 'func'])

    def test_config_iterator(self):
        test_dict = {'user': {
            'hello' : 'world'
            },
            'func': {
                'hello' : 'world'
            }
        }

        test_conf = Rickle(test_dict)

        for k in test_conf:
            self.assertEquals(k.hello, 'world')

    def test_config_has_key(self):
        test_dict = {'user': {
            'hello' : 'world'
            },
            'func': {
                'hello' : 'world'
            }
        }

        test_conf = Rickle(test_dict)

        self.assertTrue(test_conf.has('user'))
        self.assertFalse(test_conf.has('hello'))
        self.assertTrue(test_conf.has('hello', deep=True))
        self.assertFalse(test_conf.has('BYE', deep=True))
        self.assertFalse(test_conf.has('BYE', deep=False))

    def test_config_values(self):
        test_dict = {
            'user': 'BYE',
            'func': 'hello'
        }

        expected = ['BYE', 'hello']

        test_conf = BaseRickle(test_dict)

        l = test_conf.values()

        self.assertListEqual(l, expected)

    def test_config_items(self):
        test_dict = {
            'user': 'BYE',
            'func': 'hello'
        }

        test_conf = BaseRickle(test_dict)

        for k, v in test_conf.items():
            self.assertEquals(v, test_dict[k])

    def test_dump_yaml(self):
        import os

        test_dict = {'user': {
            'type': 'env',
            'load': 'USERNAME'
        },
            'func': {
                'type': 'lambda',
                'load': 'lambda x: x+1'
            }
        }

        conf = BaseRickle(test_dict)

        dumped_string = conf.to_yaml_string()

        self.assertTrue('env' in dumped_string)
        self.assertTrue('USERNAME' in dumped_string)
        self.assertTrue('lambda x: x+1' in dumped_string)

        filename = './unit_test_file_dump.yaml'

        conf.to_yaml_file(filename)

        self.assertTrue(os.path.isfile(filename))

        os.remove(filename)

    def test_dump_json(self):
        test_dict = {'user': {
            'type': 'env',
            'load': 'USERNAME'
        },
            'func': {
                'type': 'lambda',
                'load': 'lambda x: x+1'
            }
        }

        conf = BaseRickle(test_dict)

        dumped_string = conf.to_json_string()
        expected = '{"user": {"type": "env", "load": "USERNAME"}, "func": {"type": "lambda", "load": "lambda x: x+1"}}'

        self.assertEquals(dumped_string, expected)

        filename = './unit_test_file_dump.json'

        conf.to_json_file(filename)

        self.assertTrue(os.path.isfile(filename))

        os.remove(filename)

    def test_extended_config_add_function(self):
        import math
        test_conf = Rickle()

        load = """
def tester(x, c):
    y = x * 2 + c
    return math.cos(y)
        """

        args = {
            'x' : 0.42,
            'c' : 1.7
        }

        imports = ['math']

        test_conf.add_function('tester',load, args, imports)

        y = test_conf.tester(x=0.66, c=1.6)

        y_true = math.cos(0.66 * 2 + 1.6)

        self.assertEquals(y, y_true)

    def test_extended_config_add_lambda(self):
            from datetime import datetime as dd
            test_conf = Rickle()

            load = "lambda: dd.utcnow().strftime('%Y-%m-%d')"

            imports = ['from datetime import datetime as dd']

            test_conf.add_lambda('date_str', load, imports)

            y = test_conf.date_str()

            y_true = dd.utcnow().strftime('%Y-%m-%d')

            self.assertEquals(y, y_true)

    def test_pickle_rick_dict_decon(self):
        test_conf_yaml = Rickle('./tests/placebos/test_config.yaml', deep=True, load_lambda=True)

        d = test_conf_yaml.dict()

        s = test_conf_yaml.to_yaml_string()

    def test_pickle_rick_load_param(self):
        test_conf_yaml = Rickle('./tests/placebos/test_config.yaml', deep=True, arg_name='hallo_wereld', load_lambda=True)

        d = test_conf_yaml.dict()

        s = test_conf_yaml.to_yaml_string()

        s = test_conf_yaml.to_yaml_string(serialised=False)

        test_conf_yaml = Rickle('./tests/placebos/test_config.yaml', deep=True,
                                load_lambda=True)

        d = test_conf_yaml.dict()

        s = test_conf_yaml.to_yaml_string()

    def test_multi_file_load(self):
        files = ['./tests/placebos/test_config.yaml', './tests/placebos/test_second.yaml']
        test_conf_yaml = Rickle(files, deep=True, load_lambda=True)

        self.assertTrue(True)

    def test_class_definition(self):

        files = ['./tests/placebos/test_config.yaml', './tests/placebos/test_second.yaml']
        test_conf_yaml = Rickle(files, deep=True, load_lambda=True)

        obj = test_conf_yaml.TesterClass()
        obj.datenow()
        obj.math_e(99, 99)
        obj.math_e()

        self.assertTrue(True)

    def test_load_dump_load(self):
        files = ['./tests/placebos/test_config.yaml', './tests/placebos/test_second.yaml']
        test_conf_yaml = Rickle(files, deep=True, load_lambda=True)

        test_conf_yaml.to_yaml_file('./test_out.yaml')

        test_conf_yaml_reload = Rickle('./test_out.yaml', deep=True, load_lambda=True)

        test_conf_yaml_reload.BASICS.outer_math_e()

        obj = test_conf_yaml_reload.TesterClass()
        obj.datenow()
        obj.math_e(99, "^99>")

        os.remove('./test_out.yaml')

        self.assertTrue(True)

    def test_pickle_rick_dict_decon_deserialised_vs_serialised(self):
        test_conf_yaml = Rickle('./tests/placebos/test_config.yaml', deep=True, load_lambda=True)

        d = test_conf_yaml.dict()

        s = test_conf_yaml.to_yaml_string()

        d_ = test_conf_yaml.dict(serialised=True)

        s_ = test_conf_yaml.to_yaml_string(serialised=False)

        self.assertTrue(True)

    def test_load_from_url(self):
        url = 'https://official-joke-api.appspot.com/random_joke'

        r = Rickle(url)

        self.assertTrue(True)


from .__version__ import __version__, __date__
import os
import json
import copy
import warnings
from typing import Union, TypeVar
from io import TextIOWrapper
import yaml
import base64
import types
import re
import inspect
from functools import partial
import uuid

try:
    import requests
except ModuleNotFoundError as exc:
    warnings.warn(f"The module requests is not installed. This will break API calls.")

class ObjectRickler:
    """
    A class to convert Python objects to Rickle objects, deconstruct objects, create objects from Rickle objects.

    Notes:
        - `tuple` types are deconstructed as lists

    """
    def __init__(self):
        self.pat = re.compile(r'^( )*')

    def __destruct(self, value, name=None):
        if type(value) in (int, float, bool, str):
            return value

        if isinstance(value, list) or isinstance(value, tuple):
            new_list = list()
            for v in value:
                new_list.append(self.__destruct(v))
            return new_list

        if isinstance(value, dict):
            new_dict = dict()
            for k, v in value.items():
                new_dict.update({k : self.__destruct(v)})
            return new_dict

        if type(value) in (bytes, bytearray):
            return {
                'type': 'base64',
                'load': str(base64.b64encode(value))
            }

        if inspect.ismethod(value) or inspect.isfunction(value):
            signature = inspect.signature(value)
            args = dict()
            for k, v in dict(signature.parameters).items():
                if repr(v.default) == "<class 'inspect._empty'>":
                    default = None
                else:
                    default = v.default
                args.update({
                    k: default
                })

            if len(args) == 0:
                args = None

            source_lines = inspect.getsourcelines(value)[0]
            match = self.pat.match(source_lines[0])
            s = match.group(0)
            length = len(s)

            source = source_lines[0][length:]

            for s in source_lines[1:]:
                source = f'{source}{s[length:]}'


            return {
                'type': 'function',
                'name': name,
                'load': source,
                'args': args,
                'is_method' : inspect.ismethod(value)
            }

        return self.deconstruct(value)

    def to_rickle(self, obj, deep : bool = False, load_lambda : bool = False):
        """
        Transforms a Python object into a Rickle.

        Args:
            obj: Any initialised Python object.
            deep (bool): Internalize dictionary structures in lists (default = False).
            load_lambda (bool): Load python code as code or strings (default = False).

        Returns:
            Rickle: A constructed Rickle object.
        """
        d = self.deconstruct(obj)
        return Rickle(d, deep=deep, load_lambda=load_lambda)

    T = TypeVar('T')

    def from_rickle(self, rickle, cls: T, **args) -> T:
        """
        Takes a Rickle and initialises the class and updates attributes with the ones from the Rickle.

        Args:
            rickle (Rickle): Rickle to create from.
            cls (type): The class to initialise from.

        Returns:
            object: Initiliased `cls`.
        """
        if len(args) > 0:
            obj = cls(**args)
        else:
            obj = cls()

        for name, value in rickle.dict().items():
            if isinstance(value, dict) and 'type' in value.keys():
                if value['type'] == 'function':
                    _name = value.get('name', name)
                    _load = value['load']
                    _args = value.get('args', None)
                    _import = value.get('import', None)
                    f = rickle.add_function(name=_name, load=_load, args=_args, imports=_import,
                                            return_function=True, is_method=True)

                    obj.__dict__.update({_name: partial(f, obj)})
                continue
            obj.__dict__.update({name:value})

        return obj

    def deconstruct(self, obj, include_imports : bool = False, include_class_source : bool = False):
        """
        Takes (almost) any Python object and deconstructs it into a dict.

        Args:
            obj: Any object.
            include_imports (bool): Add a list of modules to import as is imported in current env (default = False).
            include_class_source (bool): Add the source of the object's class (default = False).

        Returns:
            dict: Deconstructed object in typical Rickle dictionary format.
        """
        d = dict()

        if include_class_source:
            source_lines = inspect.getsource(obj.__class__)
            d['class_source'] = {
                'type': 'class_source',
                'load' : source_lines
            }

        if include_imports:
            imports = list()

            for name, val in globals().items():
                if isinstance(val, types.ModuleType):
                    imports.append(val.__name__)

            if len(imports) > 0:
                d['python_modules'] = {
                    'type' : 'module_import',
                    'import' : imports
                }

        for name in dir(obj):
            if name.startswith('__'):
                continue
            value = getattr(obj, name)

            d[name] = self.__destruct(value, name)

        return d

    def to_yaml_string(self, obj):
        """
        Dumps the object to string.

        Args:
            obj: Any object.

        Returns:
            str: Dumped object.
        """
        d = self.deconstruct(obj)
        return yaml.safe_dump(d, None)

    def to_yaml_file(self, file_path, obj):
        """
        Dumps the object to file.

        Args:
            file_path: Filename.
            obj: Any object.
        """
        d = self.deconstruct(obj)
        with open(file_path, 'w', encoding='utf-8') as fs:
            yaml.safe_dump(d, fs)

    def to_json_string(self, obj):
        """
        Dumps the object to string.

        Args:
            obj: Any object.

        Returns:
            str: Dumped object.
        """
        d = self.deconstruct(obj)
        return json.dumps(d)

    def to_json_file(self, file_path, obj):
        """
        Dumps the object to file.

        Args:
            file_path: Filename.
            obj: Any object.
        """
        d = self.deconstruct(obj)
        with open(file_path, 'w', encoding='utf-8') as fs:
            json.dump(d, fs)

class BaseRickle:
    """
        A base class that creates internal structures from embedded structures.

        Args:
            base (str,dict,TextIOWrapper, list): String (YAML or JSON, file path to YAML/JSON file, URL), text IO stream, dict (default = None).
            deep (bool): Internalize dictionary structures in lists (default = False).
            strict (bool): Check keywords, if YAML/JSON key is Rickle keyword (or member of object) raise ValueError (default = True).
            **init_args (kw_args): Additional arguments for string replacement

        Raises:
            ValueError: If the given base object can not be handled. Also raises if YAML key is already member of Rickle.
    """
    def _iternalize(self, dictionary : dict, deep : bool, **init_args):
        for k, v in dictionary.items():
            self._check_kw(k)
            if isinstance(v, dict):
                self.__dict__.update({k: BaseRickle(base=v, deep=deep, strict=self.__strict, **init_args)})
                continue
            if isinstance(v, list) and deep:
                new_list = list()
                for i in v:
                    if isinstance(i, dict):
                        new_list.append(BaseRickle(base=i, deep=deep, strict=self.__strict, **init_args))
                    else:
                        new_list.append(i)
                self.__dict__.update({k: new_list})
                continue

            self.__dict__.update({k:v})

    def __init__(self, base : Union[dict,str,TextIOWrapper,list] = None, deep : bool = False, strict: bool = True, **init_args):
        self.__meta_info = dict()
        self.__strict = strict
        stringed = ''
        if base is None:
            return
        if isinstance(base, dict):
            self._iternalize(base, deep, **init_args)
            return

        if isinstance(base, TextIOWrapper):
            stringed = base.read()
        elif isinstance(base, list):
            for file in base:
                with open(file, 'r') as f:
                    stringed = f'{stringed}\n{f.read()}'
        elif os.path.isfile(base):
            with open(base, 'r') as f:
                stringed = f.read()
        elif isinstance(base, str):
            try:
                from urllib3.util import parse_url
                try:
                    parsed = parse_url(base)
                    if all([parsed.scheme, parsed.host]):
                        response = requests.get(url=base,)
                        dict_data = response.json()
                        self._iternalize(dict_data, deep, **init_args)
                        return
                except:
                    pass
            except (ImportError, ModuleNotFoundError):
                pass

            stringed = base

        if not init_args is None:
            for k, v in init_args.items():
                _k = f'_|{k}|_'
                stringed = stringed.replace(_k, json.dumps(v))

        try:
            dict_data = yaml.safe_load(stringed)
            self._iternalize(dict_data, deep, **init_args)
            return
        except Exception as exc:
            print("Tried YAML: {}".format(exc))
        try:
            dict_data = json.loads(base)
            self._iternalize(dict_data, deep, **init_args)
            return
        except Exception as exc:
            print("Tried JSON: {}".format(exc))


        raise ValueError('Base object could not be internalized, type {} not handled'.format(type(base)))

    def __repr__(self):
        keys = self.__dict__
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys if not str(k).__contains__(self.__class__.__name__) and not str(k).endswith('__meta_info') )
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __str__(self):
        return self.to_yaml_string()

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __len__(self):
        return len(self.__dict__)

    def __iter__(self):
        self.__n = 0
        return self

    def __next__(self):
        current_loop = 0
        if self.__n < len(self.__dict__):
            name = list(self.__dict__.keys())[self.__n]
            while self.__eval_name(name) and current_loop < 9:
                self.__n += 1
                current_loop += 1
                if self.__n < len(self.__dict__):
                    name = list(self.__dict__.keys())[self.__n]
                else:
                    raise StopIteration
            if self.__n < len(self.__dict__):
                obj = self.__dict__[list(self.__dict__.keys())[self.__n]]
                self.__n += 1
                return obj
            else:
                raise StopIteration
        else:
            raise StopIteration

    def __getitem__(self, key):
        if key is None:
            raise KeyError("NoneType is not a valid key type")
        if not isinstance(key, str):
            raise TypeError("Key can only be of case sensitive string type")

        return self.__dict__[key]

    def __setitem__(self, key, value):
        if key is None:
            raise KeyError("NoneType is not a valid key type")
        if not isinstance(key, str):
            raise TypeError("Key can only be of case sensitive string type")

        self.__dict__.update({key: value})

    def __delitem__(self, key):
        if key is None:
            raise KeyError("NoneType is not a valid key type")
        if not isinstance(key, str):
            raise TypeError("Key can only be of case sensitive string type")

        del self.__dict__[key]

    def __search_path(self, key, dictionary=None, parent_path=None):
        if dictionary is None:
            dictionary = self.dict()
        if parent_path is None:
            parent_path = ''
        values = list()
        if key in dictionary:
            values = [f'{parent_path}/{key}']
        for k, v in dictionary.items():
            if isinstance(v, BaseRickle):
                try:
                    value = self.__search_path(key=key, dictionary=v.dict(),  parent_path=f'{parent_path}/{k}')
                    values.extend(value)
                except StopIteration:
                    continue
            if isinstance(v, dict):
                try:
                    value = self.__search_path(key=key, dictionary=v,  parent_path=f'{parent_path}/{k}')
                    values.extend(value)
                except StopIteration:
                    continue
        if len(values) > 0:
            return values
        raise StopIteration

    def search_path(self, key : str) -> list:
        """
        Search the current Rickle for all paths that match the search key. Returns empty list if nothing is found.

        Args:
            key (str): The key to search.

        Returns:
            list: all paths found.
        """
        try:
            return self.__search_path(key=key)
        except StopIteration:
            return list()


    def __call__(self, path : str, **kwargs):
        """
        Rickle objects can be queried via a path string.

        Notes:
            '/' => root.
            '/name' => member.
            '/path/to/name?param=1' => lambda/function.
            If '?' is in path the inline parameters are used and kwargs are ignored.

        Args:
            path (str): The path as a string, down to the last mentioned node.

        Returns:
            Any: Value of node of function.
        """

        if not path.startswith('/'):
            raise KeyError('Missing root path /')
        if path == '/':
            return self

        path_list = path.split('/')

        current_node = self

        for node_name in path_list[1:]:
            if '?' in node_name:
                node_name = node_name.split('?')[0]
            current_node = current_node.get(node_name)
            if current_node is None:
                raise NameError(f'The path {path} could not be traversed')

        if '?' in path_list[-1]:
            import ast
            args_string = path_list[-1].split('?')[-1]
            args = {a.split('=')[0] : a.split('=')[1] for a in args_string.split('&')}
            type_guessed_args = dict()
            for n, v in args.items():
                v_stripped = v.strip()
                try:
                    literal = ast.literal_eval(v_stripped)
                except Exception as exc:
                    raise TypeError(f"Could not guess the parameter type, {exc}")

                if isinstance(literal, str):
                    # This is intentionally done to strip away the literal quotes
                    type_guessed_args[n] = v_stripped[1:-1]
                else:
                    type_guessed_args[n] = literal
            try:
                return current_node(**type_guessed_args)
            except Exception as exc:
                raise TypeError(f'{exc} occurred. The node in the path {path} is of type {type(current_node)} or does not match the query')

        if inspect.isfunction(current_node):
            try:
                return current_node(**kwargs)
            except Exception as exc:
                raise TypeError(f'{exc} occurred. The node in the path {path} is of type {type(current_node)} or does not match the query')

        else:
            return current_node


    def __eval_name(self, name):
        if str(name).__contains__(self.__class__.__name__) or str(name).endswith('__n'):
            return True
        else:
            return False

    def _check_kw(self, name):
        if self.__strict and name in dir(self):
            raise ValueError(f"Unable to add key '{name}', reserved keyword in Rickle. Use strict=False.")

    def _recursive_search(self, dictionary, key):
        if key in dictionary:
            return dictionary[key]
        for k, v in dictionary.items():
            if isinstance(v, BaseRickle):
                try:
                    value = self._recursive_search(v.__dict__, key)
                    return value
                except StopIteration:
                    continue
            if isinstance(v, dict):
                try:
                    value = self._recursive_search(v, key)
                    return value
                except StopIteration:
                    continue
        raise StopIteration

    def items(self):
        """
        Iterate through all key value pairs.

        Yields:
            tuple: str, object.
        """
        d = self.dict()
        for key in d.keys():
            if self.__eval_name(key):
                continue
            yield key, d[key]

    def get(self, key : str, default=None, do_recursive : bool = False):
        """
        Acts as a regular get from a dictionary but can employ a recursive search of structure and returns the first found key-value pair.

        Note:
            Document paths like '/root/to/path' can also be used. If the path can not be traversed, the default value is returned.

        Args:
            key (str): key string being searched.
            default (any): Return value if nothing is found.
            do_recursive (bool): Search recursively until first match is found (default = False).

        Returns:
            obj: value found, or None for nothing found.
        """
        try:
            if '/' in key:
                v = self(key)
                return v
            if do_recursive:
                value = self._recursive_search(self.__dict__, key)
            else:
                value = self.__dict__.get(key, default)
            return value
        except StopIteration:
            return default
        except NameError:
            return default
        except Exception as ex:
            raise ex

    def set(self, key: str, value):
        """
        As with the `get` method, this method can be used to update the inherent dictionary with new values.

        Note:
            Document paths like '/root/to/path' can also be used. If the path can not be traversed, an error is raised.

        Args:
            key (str): key string to set.
            value: Any Python like value that can be deserialised.
        """

        if '/' in key and not key.startswith('/'):
            raise KeyError('Missing root path /')
        if not '/' in key:
            key = f"/{key}"

        if key == '/':
            raise NameError('Can not set a value to self')

        path_list = key.split('/')

        current_node = self

        for node_name in path_list[1:-1]:
            current_node = current_node.get(node_name)
            if current_node is None:
                raise NameError(f'The path {key} could not be traversed')

        if '?' in path_list[-1]:
            raise KeyError(f'Function params "{path_list[-1]}" included in path!')

        current_node.__dict__.update({path_list[-1]: value})

    def remove(self, key: str):
        if '/' in key and not key.startswith('/'):
            raise KeyError('Missing root path /')
        if not '/' in key:
            key = f"/{key}"

        if key == '/':
            raise NameError('Can not remove self')

        path_list = key.split('/')

        current_node = self

        for node_name in path_list[1:-1]:
            current_node = current_node.get(node_name)
            if current_node is None:
                raise NameError(f'The path {key} could not be traversed')

        if '?' in path_list[-1]:
            raise KeyError(f'Function params "{path_list[-1]}" included in path!')

        del current_node.__dict__[path_list[-1]]


    def values(self):
        """
        Gets the higher level values of the current Rick object.

        Returns:
            list: of objects.
        """
        d = self.dict()
        keys = list(d.keys())
        objects = [d[k] for k in keys if not self.__eval_name(k)]

        return objects

    def keys(self):
        """
        Gets the higher level keys of the current Rick object.

        Returns:
            list: of keys.
        """
        d = self.dict()
        keys = list(d.keys())
        keys = [k for k in keys if not self.__eval_name(k)]

        return keys

    def dict(self, serialised : bool = False):
        """
        Deconstructs the whole object into a Python dictionary.

        Args:
            serialised (bool): Give a Python dictionary in serialised (True) form or deserialised (default = False).

        Notes:
            Functions and lambdas are always given in serialised form.

        Returns:
            dict: of object.
        """
        d = dict()
        for key, value in self.__dict__.items():
            if self.__eval_name(key) or str(key).endswith('__meta_info'):
                continue
            if isinstance(value, BaseRickle) or isinstance(value, Rickle):
                d[key] = value.dict(serialised=serialised)
            elif isinstance(value, list):
                new_list = list()
                for element in value:
                    if isinstance(element, BaseRickle):
                        new_list.append(element.dict(serialised=serialised))
                    else:
                        new_list.append(element)
                d[key] = new_list
            else:
                d[key] = value
        return d

    def has(self, key : str, deep=False) -> bool:
        """
        Checks whether the key exists in the object.

        Args:
            key (str): key string being searched.
            deep (bool): whether to search deeply (default = False).

        Returns:
            bool: if found.
        """
        if key in self.dict():
            return True
        if deep:
            try:
                self._recursive_search(self.dict(), key)
                return True
            except StopIteration:
                return False
        return False

    def to_yaml_file(self, file_path : str, serialised : bool = True):
        """
        Does a self dump to a YAML file.

        Args:
            file_path (str): File path.
            serialised (bool): Give a Python dictionary in serialised (True) form or deserialised (default = True).

        Notes:
            Functions and lambdas are always given in serialised form.
        """
        self_as_dict = self.dict(serialised=serialised)
        with open(file_path, 'w', encoding='utf-8') as fs:
            yaml.safe_dump(self_as_dict, fs)

    def to_yaml_string(self, serialised : bool = True):
        """
        Dumps self to YAML string.

        Args:
            serialised (bool): Give a Python dictionary in serialised (True) form or deserialised (default = True).

        Notes:
            Functions and lambdas are always given in serialised form.

        Returns:
            str: YAML representation.
        """
        self_as_dict = self.dict(serialised=serialised)
        return yaml.safe_dump(self_as_dict, None)

    def to_json_file(self, file_path: str, serialised : bool = True):
        """
        Does a self dump to a JSON file.

        Args:
            file_path (str): File path.
            serialised (bool): Give a Python dictionary in serialised (True) form or deserialised (default = True).

        Notes:
            Functions and lambdas are always given in serialised form.
        """
        self_as_dict = self.dict(serialised=serialised)
        with open(file_path, 'w', encoding='utf-8') as fs:
            json.dump(self_as_dict, fs)

    def to_json_string(self, serialised : bool = True):
        """
        Dumps self to YAML string.

        Args:
            serialised (bool): Give a Python dictionary in serialised (True) form or deserialised (default = True).

        Notes:
            Functions and lambdas are always given in serialised form.

        Returns:
            str: JSON representation.
        """
        self_as_dict = self.dict(serialised=serialised)
        return json.dumps(self_as_dict)

    def meta(self, name):
        """
        Get the metadata for a property.

        Args:
            name (str): The name of the proprty.

        Returns:
            dict: The metadata as a dict.
        """
        return self.__meta_info[name]

    def add_attr(self, name, value):
        """
        Add a new attribute member to Rick.

        Args:
            name (str): Property name.
            value (any): Value of new member.
        """
        self._check_kw(name)
        self.__dict__.update({name: value})
        self.__meta_info[name] = {'type': 'attr', 'value': value}

class Rickle(BaseRickle):
    """
        An extended version of the BasicRick that can load OS environ variables and Python functions.

        Args:
            base (str, list): String (YAML or JSON, file path to YAML/JSON file) or list of file paths, text IO stream, dict.
            deep (bool): Internalize dictionary structures in lists.
            load_lambda (bool): Load lambda as code or strings.
            strict (bool): Check keywords, if YAML/JSON key is Rickle keyword (or member of object) raise ValueError (default = True).
            **init_args (kw_args): Additional arguments for string replacement

        Raises:
            ValueError: If the given base object can not be handled. Also raises if YAML key is already member of Rickle.
    """

    def _iternalize(self, dictionary: dict, deep: bool, **init_args):
        for k, v in dictionary.items():
            self._check_kw(k) # Redundant but easier to check twice than to paste 10 times
            if isinstance(v, dict):
                if 'type' in v.keys():
                    if v['type'] == 'env':
                        self.add_env_variable(name=k,
                                              load=v['load'],
                                              default=v.get('default', None))
                        continue
                    if v['type'] == 'base64':
                        self.add_base64(name=k,
                                              load=v['load'])
                        continue
                    if v['type'] == 'module_import':
                        self.add_module_import(name=k,
                                              imports=v['import'])
                        continue
                    if v['type'] == 'from_file':
                        self.add_from_file(name=k,
                                           file_path=v['file_path'],
                                           load_as_rick=v.get('load_as_rick', False),
                                           deep=v.get('deep', False),
                                           load_lambda=v.get('load_lambda', False),
                                           is_binary=v.get('is_binary', False),
                                           encoding=v.get('encoding', 'utf-8'),
                                           hot_load=v.get('hot_load', False))
                        continue
                    if v['type'] == 'from_csv':
                        self.add_csv_file(name=k,
                                          file_path=v['file_path'],
                                          fieldnames=v.get('fieldnames', None),
                                          load_as_rick=v.get('load_as_rick', False),
                                          encoding=v.get('encoding', 'utf-8'))
                        continue
                    if v['type'] == 'api_json':
                        self.add_api_json_call(name=k,
                                               url=v['url'],
                                               http_verb=v.get('http_verb', 'GET'),
                                               headers=v.get('headers', None),
                                               params=v.get('params', None),
                                               body=v.get('body', None),
                                               load_as_rick=v.get('load_as_rick', False),
                                               load_lambda=v.get('load_lambda', False),
                                               deep=v.get('deep', False),
                                               expected_http_status=v.get('expected_http_status', 200),
                                               hot_load=v.get('hot_load', False))
                        continue
                    if v['type'] == 'html_page':
                        self.add_html_page(name=k,
                                           url=v['url'],
                                           headers=v.get('headers', None),
                                           params=v.get('params', None),
                                           expected_http_status=v.get('expected_http_status', 200),
                                           hot_load=v.get('hot_load', False))
                        continue
                    if v['type'] == 'lambda':
                        load = v['load']
                        imports = v.get('import', None)
                        safe_load = os.getenv("RICKLE_SAFE_LOAD", None)
                        if init_args and init_args['load_lambda'] and safe_load is None:
                            self.add_lambda(name=k,
                                            load=load,
                                            imports=imports)
                        else:
                            self.__dict__.update({k: v})
                        continue
                    if v['type'] == 'class_definition':
                        name = v.get('name', k)
                        attributes = v['attributes']
                        imports = v.get('import', None)
                        safe_load = os.getenv("RICKLE_SAFE_LOAD", None)
                        if init_args and init_args['load_lambda'] and safe_load is None:
                            self.add_class_definition(name=name,
                                              attributes=attributes,
                                              imports=imports)
                        else:
                            self.__dict__.update({k: v})
                        continue
                    if v['type'] == 'function':
                        name = v.get('name', k)
                        load = v['load']
                        args_dict = v.get('args', None)
                        imports = v.get('import', None)
                        is_method = v.get('is_method', False)

                        safe_load = os.getenv("RICKLE_SAFE_LOAD", None)
                        if init_args and init_args['load_lambda'] and safe_load is None:
                            self.add_function(name=name,
                                              load=load,
                                              args=args_dict,
                                              imports=imports,
                                              is_method=is_method)
                        else:
                            self.__dict__.update({k: v})
                        continue

                self.__dict__.update({k:Rickle(base=v, deep=deep, strict=self.__strict, **init_args)})
                continue
            if isinstance(v, list) and deep:
                new_list = list()
                for i in v:
                    if isinstance(i, dict):
                        new_list.append(Rickle(base=i, deep=deep, strict=self.__strict, **init_args))
                    else:
                        new_list.append(i)
                self.__dict__.update({k: new_list})
                continue
            self.__dict__.update({k: v})

    def __init__(self, base: Union[dict, str, TextIOWrapper, list] = None,
                 deep: bool = False,
                 load_lambda: bool = False,
                 strict: bool = True,
                 **init_args):
        self.__meta_info = dict()
        init_args['load_lambda'] = load_lambda
        init_args['deep'] = deep
        init_args['strict'] = strict
        self.__strict = strict
        self.__init_args = init_args
        super().__init__(base, **init_args)

    def __eval_name(self, name):
        if str(name).__contains__(self.__class__.__name__) or str(name).endswith('__n'):
            return True
        else:
            return False

    def meta(self, name):
        """
        Get the metadata for a property.

        Args:
            name (str): The name of the proprty.

        Returns:
            dict: The metadata as a dict.
        """
        return self.__meta_info[name]

    def dict(self, serialised : bool = False):
        """
        Deconstructs the whole object into a Python dictionary.

        Args:
            serialised (bool): Give a Python dictionary in serialised (True) form or deserialised (default = False).

        Notes:
            Functions and lambdas are always given in serialised form.

        Returns:
            dict: of object.
        """
        d = dict()
        for key, value in self.__dict__.items():
            if self.__eval_name(key):
                continue
            if serialised and key in self.__meta_info.keys():
                d[key] = self.__meta_info[key]
            elif key in self.__meta_info.keys() and \
                    self.__meta_info[key]['type'] in ['function', 'lambda', 'class_definition']:
                d[key] = self.__meta_info[key]
            elif key in self.__meta_info.keys() and \
                    self.__meta_info[key]['type'] in ['from_file', 'html_page', 'api_json'] and \
                    self.__meta_info[key]['hot_load']:
                d[key] = self.__meta_info[key]
            elif isinstance(value, BaseRickle):
                d[key] = value.dict(serialised=serialised)
            elif isinstance(value, list):
                new_list = list()
                for element in value:
                    if isinstance(element, BaseRickle):
                        new_list.append(element.dict(serialised=serialised))
                    else:
                        new_list.append(element)
                d[key] = new_list
            else:
                d[key] = value
        return d

    def add_module_import(self, name, imports : list):
        """
        Add global Python module imports.

        Args:
            name: Name of import list.
            imports (list): List of strings of Python module names.

        """
        self._check_kw(name)
        for i in imports:
            if 'import' in i:
                exec(i, globals())
            else:
                exec('import {}'.format(i), globals())

        self.__meta_info[name] = {'type' : 'module_import', 'import' : imports}

    def add_function(self, name, load, args : dict = None, imports : list = None,
                     return_function : bool = False,
                     is_method : bool = False):
        """
        Add a new function to Rick.

        Args:
            name (str): Property name.
            load (str): Python code containing the function.
            args (dict): Key-value pairs of arguments with default values (default = None).
            imports (list): Python modules to import (default = None).
            return_function (bool): Add to rickle or return the function (default = False).
            is_method (bool): Indicates whether class method source includes `self` (default = False).

        Examples:
            Basic example for adding to a PickleRick:
                >> test_rick = PickleRick()

                >> load = '''
                        def tester(x, c):
                            y = x * 2 + c
                            return math.cos(y)
                        '''

                >> args = { 'x' : 0.42, 'c' : 1.7 }

                >> imports = ['math']

                >> test_rick.add_function('tester',load, args, imports)

                >> y = test_rick.tester(x=0.66, c=1.6)

        """
        if not return_function:
            self._check_kw(name)
        if imports and isinstance(imports, list):
            for i in imports:
                if 'import' in i:
                    exec(i, globals())
                else:
                    exec('import {}'.format(i), globals())
        suffix  = str(uuid.uuid4().hex)

        _load = load.replace(f'def {name}(', f'def {name}{suffix}(')
        exec(_load, globals())
        if args and isinstance(args, dict):
            if is_method:
                arg_list = ['self=self']
                arg_list_defaults = ['self=self']
            else:
                arg_list = list()
                arg_list_defaults = list()
            for arg in args.keys():
                default_value = args[arg]
                if isinstance(default_value, str):
                    arg_list_defaults.append(f"{arg}='{default_value}'")
                else:
                    arg_list_defaults.append(f"{arg}={default_value}")
                arg_list.append(f"{arg}={arg}")

            func_string = 'lambda {args_default}: {name}({args})'.format(
                args_default=','.join(arg_list_defaults),
                args=','.join(arg_list),
                name=name+suffix)
        else:
            if is_method:
                func_string = 'lambda self: {name}(self)'.format(name=name+suffix)
            else:
                func_string = 'lambda: {name}()'.format(name=name+suffix)

        if return_function:
            return eval(func_string)

        self.__dict__.update({name: eval(func_string)})
        self.__meta_info[name] = {'type' : 'function', 'name' : name, 'args' : args, 'import' : imports,
                                  'load' : load, 'is_method' : is_method}

    def add_lambda(self, name, load, imports : list = None, return_lambda : bool = False, is_method : bool = False):
        """
        Add a Python lambda to Rick.

        Args:
            name (str): Property name.
            load (str): Python code containing the lambda.
            imports (list): Python modules to import (default = None).
            return_lambda (bool): Add to rickle or return the lambda (default = False).
            is_method (bool): Add `self` param (default = False).

        Examples:
            Basic example for adding to a PickleRick:
                >> test_rick = PickleRick()

                >> load = "lambda: dd.utcnow().strftime('%Y-%m-%d')"

                >> imports = ['from datetime import datetime as dd']

                >> test_rick.add_lambda('date_str', load, imports)

                >> date_string = test_rick.date_str()
        """
        if not return_lambda:
            self._check_kw(name)
        if imports and isinstance(imports, list):
            for i in imports:
                if 'import' in i:
                    exec(i, globals())
                else:
                    exec('import {}'.format(i), globals())

        if load.startswith('lambda'):
            _load = load
        elif is_method:
                _load = f'lambda self: {load}'
        else:
            _load = f'lambda: {load}'

        if return_lambda:
            return eval(_load)

        self.__dict__.update({name: eval(_load)})
        self.__meta_info[name] = {'type' : 'lambda', 'import' : imports, 'load' : load}

    def add_env_variable(self, name, load, default = None):
        """
        Add a new OS ENVIRONMENT VARIABLE to Rick.

        Args:
            name (str): Property name.
            load (str): ENV var name.
            default (any): Default to value (default = None).
        """
        self._check_kw(name)
        self.__dict__.update({name: os.getenv(load, default)})
        self.__meta_info[name] = {'type' : 'env', 'load' : load, 'default' : default}

    def add_base64(self, name, load):
        """
        Add Base 64 encoded byte string data.

        Args:
            name (str): Property name.
            load (str): Base 64 encoded data.
        """
        self._check_kw(name)
        b = base64.b64decode(load)
        self.__dict__.update({name: b})
        self.__meta_info[name] = {'type': 'base64',
                                  'load' : load
                                  }

    def add_csv_file(self,
                     name,
                     file_path : str,
                     fieldnames : list = None,
                     load_as_rick  : bool = False,
                     encoding : str = 'utf-8'
                     ):
        """
        Adds the ability to load CSV data as lists or even a list of Ricks where the column names are the properties.

        Args:
            name (str): Property name.
            file_path (str): File path to load from.
            fieldnames (list): Column headers (default = None).
            load_as_rick (bool): If true, loads and creates Rick from source, else loads the contents as text (default = False).
            encoding (str): If text, encoding can be specified (default = 'utf-8').

        """
        self._check_kw(name)
        import csv
        with open(file_path, 'r', encoding=encoding) as file:
            dialect = csv.Sniffer().sniff(file.read(1024))
            file.seek(0)
            l = list()

            if load_as_rick:
                csv_file = csv.DictReader(file, fieldnames=fieldnames, dialect=dialect)

                for row in csv_file:
                    l.append(dict(row))

                self._iternalize({name: l}, deep=True)
            elif not fieldnames is None:

                columns = {c : list() for c in fieldnames}

                csv_file = csv.DictReader(file, fieldnames=fieldnames, dialect=dialect)

                for row in csv_file:
                    for k,v in row.items():
                        columns[k].append(v)

                self._iternalize({name: columns}, deep=False)
            else:
                csv_file = csv.reader(file, dialect=dialect)

                for row in csv_file:
                    l.append(row)

                self.__dict__.update({name: l})

        self.__meta_info[name] = {'type': 'from_csv',
                                  'file_path': file_path,
                                  'load_as_rick': load_as_rick,
                                  'fieldnames' : fieldnames,
                                  'encoding': encoding
                                  }

    def add_dataframe(self):
        # Implement later
        # self._check_kw(name)
        raise NotImplementedError()

    def _load_from_file(self,
                      file_path: str,
                      load_as_rick: bool = False,
                      deep: bool = False,
                      load_lambda: bool = False,
                      is_binary: bool = False,
                      encoding: str = 'utf-8'):
        if load_as_rick and not is_binary:
            args = copy.copy(self.__init_args)
            args['load_lambda'] = load_lambda
            args['deep'] = deep
            return Rickle(file_path, **args)
        else:
            if is_binary:
                with open(file_path, 'rb') as fn:
                    return fn.read()
            else:
                with open(file_path, 'r', encoding=encoding) as fn:
                    return fn.read()

    def add_from_file(self, name,
                      file_path : str,
                      load_as_rick : bool = False,
                      deep : bool = False,
                      load_lambda : bool = False,
                      is_binary : bool = False,
                      encoding : str = 'utf-8',
                      hot_load : bool = False):
        """
        Adds the ability to further load Ricks from other YAML or JSON files, or alternatively load a text file.
        This opens up dynamic possibility, but with that it also opens up extreme security vulnerabilities.
        Only ever load files from trusted sources.
        **Important note: Even with ``deep`` and ``load_lambda`` set to False, further file or API calls could be found within the source that loads lambda functions.**
        **Important note: Be careful to never self-reference a file, i.e. don't load the same file from within itself to avoid infinte looping.**

        Args:
            name (str): Property name.
            file_path (str): File path to load from.
            load_as_rick (bool): If true, loads and creates Rick from source, else loads the contents as text (default = False).
            deep (bool): Internalize dictionary structures in lists (default = False).
            load_lambda (bool): Load lambda as code or strings (default = False).
            is_binary (bool): If the file is a binary file (default = False).
            encoding (str): If text, encoding can be specified (default = 'utf-8').
            hot_load (bool): Load the data on calling or load it only once on start (cold) (default = False).
        """
        self._check_kw(name)
        if hot_load:
            _load = f"""lambda self=self: self._load_from_file(file_path='{file_path}',
                                          load_as_rick={load_as_rick},
                                          deep={deep},
                                          load_lambda={load_lambda},
                                          is_binary={is_binary},
                                          encoding='{encoding}')"""

            self.__dict__.update({name: eval(_load)})
        else:
            result = self._load_from_file(file_path=file_path,
                                          load_as_rick=load_as_rick,
                                          deep=deep,
                                          load_lambda=load_lambda,
                                          is_binary=is_binary,
                                          encoding=encoding)

            self.__dict__.update({name: result})

        self.__meta_info[name] = {'type': 'from_file',
                                  'file_path' : file_path,
                                  'load_as_rick' : load_as_rick,
                                  'deep' : deep,
                                  'load_lambda' : load_lambda,
                                  'is_binary' : is_binary,
                                  'encoding' : encoding,
                                  'hot_load' : hot_load
                                  }

    def _load_html_page(self,
                      url : str,
                      headers : dict = None,
                      params : dict = None,
                      expected_http_status : int = 200):
        r = requests.get(url=url, params=params, headers=headers)

        if r.status_code == expected_http_status:
            return r.text
        else:
            raise ValueError(f'Unexpected HTTP status code in response {r.status_code}')

    def add_html_page(self,
                      name,
                      url : str,
                      headers : dict = None,
                      params : dict = None,
                      expected_http_status : int = 200,
                      hot_load : bool = False):
        """
        Loads HTML page as property.

        Args:
            name (str): Property name.
            url (str): URL to load from.
            headers (dict): Key-value pair for headers (default = None).
            params (dict): Key-value pair for parameters (default = None).
            expected_http_status (int): Should a none 200 code be expected (default = 200).
            hot_load (bool): Load the data on calling or load it only once on start (cold) (default = False).

        """
        self._check_kw(name)
        if hot_load:
            _load = f"""lambda self=self: self._load_html_page(url='{url}',
                                          headers={headers},
                                          params={params},
                                          expected_http_status={expected_http_status})"""

            self.__dict__.update({name: eval(_load)})
        else:
            result = self._load_html_page(url=url,
                                          headers=headers,
                                          params=params,
                                          expected_http_status=expected_http_status)

            self.__dict__.update({name: result})

        self.__meta_info[name] = {'type': 'html_page',
                                  'url': url,
                                  'headers': headers,
                                  'params': params,
                                  'expected_http_status': expected_http_status,
                                  'hot_load' : hot_load
                                  }

    def add_class_definition(self, name, attributes, imports : list = None ):
        """
        Adds a class definition, with attributes such as functions and lambdas.

        Args:
            name (str): Property name.
            attributes (dict): Standard items or Rickle function definitions.
            imports (list): Python modules to import (default = None).

        """
        self._check_kw(name)
        if imports and isinstance(imports, list):
            for i in imports:
                if 'import' in i:
                    exec(i, globals())
                else:
                    exec('import {}'.format(i), globals())

        _attributes = dict()
        for k,v in attributes.items():
            if isinstance(v, dict):
                if 'type' in v.keys() and v['type'] == 'function':
                    _name = v.get('name', k)
                    load = v['load']
                    args_dict = v.get('args', None)
                    imports = v.get('import', None)
                    is_method = v.get('is_method', False)
                    _attributes[_name] = self.add_function(name=_name,load=load,args=args_dict,imports=imports,
                                                           return_function=True, is_method=is_method)
                    continue
                if 'type' in v.keys() and v['type'] == 'lambda':
                    load = v['load']
                    imports = v.get('import', None)
                    _attributes[k] = self.add_lambda(name=k, load=load, imports=imports, return_lambda=True, is_method=True)
                    continue
            _attributes[k] = v


        self.__dict__.update({name: type(name,(), _attributes)})

        self.__meta_info[name] = {'type' : 'class_definition', 'name' : name, 'import' : imports, 'attributes' : attributes}

    def _load_api_json_call(self,
                                url : str,
                                http_verb : str = 'GET',
                                headers : dict = None,
                                params : dict = None,
                                body : dict = None,
                                load_as_rick: bool = False,
                                deep : bool = False,
                                load_lambda : bool = False,
                                expected_http_status : int = 200):
        if http_verb.lower() == 'post':
            r = requests.post(url=url, data=body, headers=headers)
        else:
            r = requests.get(url=url, params=params, headers=headers)

        if r.status_code == expected_http_status:
            json_dict = r.json()
            if load_as_rick:
                args = copy.copy(self.__init_args)
                args['load_lambda'] = load_lambda
                args['deep'] = deep

                return Rickle(json_dict, **args)
            else:
                return json_dict
        else:
            raise ValueError(f'Unexpected HTTP status code in response {r.status_code}')


    def add_api_json_call(self, name,
                          url : str,
                          http_verb : str = 'GET',
                          headers : dict = None,
                          params : dict = None,
                          body : dict = None,
                          load_as_rick: bool = False,
                          deep : bool = False,
                          load_lambda : bool = False,
                          expected_http_status : int = 200,
                          hot_load : bool = False):
        """
        Load a JSON response from a URL and create a Rick from it. This opens up dynamic possibility,
        but with that it also opens up extreme security vulnerabilities. Only ever load JSON objects from trusted sources.
        **Important note: Even with ``deep`` and ``load_lambda`` set to False, further API calls could be found within the source that loads lambda functions.**
        **Important note: Be careful to never self-reference an API call, i.e. don't load the same API from within itself to avoid infinte looping.**

        Args:
            name (str): Property name.
            url (str): URL to load from.
            http_verb (str): Either 'POST' or 'GET' allowed (default = 'GET').
            headers (dict): Key-value pair for headers (default = None).
            params (dict): Key-value pair for parameters (default = None).
            body (dict): Key-value pair for data (default = None).
            load_as_rick (bool): If true, loads and creates Rick from source, else loads the contents as dictionary (default = False).
            deep (bool): Internalize dictionary structures in lists (default = False).
            load_lambda (bool): Load lambda as code or strings (default = False).
            expected_http_status (int): Should a none 200 code be expected (default = 200).
            hot_load (bool): Load the data on calling or load it only once on start (cold) (default = False).

        """
        self._check_kw(name)
        if hot_load:
            _load = f"""lambda self=self: self._load_api_json_call(url='{url}', 
                                http_verb='{http_verb}', 
                                headers={headers}, 
                                params={params}, 
                                body={body},
                                load_as_rick={load_as_rick},
                                deep={deep},
                                load_lambda={load_lambda},
                                expected_http_status={expected_http_status})"""

            self.__dict__.update({name: eval(_load)})
        else:
            result = self._load_api_json_call(url=url,
                                               http_verb=http_verb,
                                               headers=headers,
                                               params=params,
                                               body=body,
                                               load_as_rick=load_as_rick,
                                               deep=deep,
                                               load_lambda=load_lambda,
                                               expected_http_status=expected_http_status)

            self.__dict__.update({name: result})

        self.__meta_info[name] = {'type': 'api_json',
                                  'url': url,
                                  'http_verb': http_verb,
                                  'headers' : headers,
                                  'params' : params,
                                  'body' : body,
                                  'deep' : deep,
                                  'load_lambda' : load_lambda,
                                  'expected_http_status' : expected_http_status,
                                  'hot_load' : hot_load
                                  }
